# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2024 Bryant E. McDonnell (See AUTHORS)
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
"""Base class for a SWMM Simulation."""

# Standard import
from warnings import warn

# Local imports
from pyswmm.swmm5 import PySWMM, PYSWMMException
from pyswmm.toolkitapi import SimulationTime, SimulationUnits
from pyswmm.errors import MultiSimulationError


class _SimulationStateManager(object):
    """This manager was created to be a guardrail for the PySWMM developers
    experience.  In the event the developer is unaware of the non thread-safe
    non-reenterant quality of USEPA-SWMM, this prevents the developer from trying
    to open multiple instances of SWMM inside one instance of Python.

    The State Manager also give the option to show a simulation progress bar for
    the users running this code on the command line."""

    def __init__(self):
        self._sim_is_instantiated = False

    @property
    def sim_is_instantiated(self) -> bool:
        return self._sim_is_instantiated

    @sim_is_instantiated.setter
    def sim_is_instantiated(self, val: bool) -> None:
        self._sim_is_instantiated = val


# Module level instance for the Simulation Manager.
_sim_state_instance = _SimulationStateManager()


class Simulation(object):
    """
    Base class for a SWMM Simulation.

    The model object provides several options to run a simulation.

    Initialize the Simulation class.

    :param str inpfile: Name of SWMM input file (default '')
    :param str rptfile: Report file to generate (default None)
    :param str binfile: Optional binary output file (default None)
    :param SimulationPreConfig sim_preconfig: Optional Pre Config (default None)

    Examples:

    Intialize using with statement.  This automatically cleans up
    after a simulation

    .. code-block:: python

        from pyswmm import Simulation

        with Simulation('tests/data/model_weir_setting.inp') as sim:
            for step in sim:
                pass

    Initialize the simulation and execute. This style does not allow
    the user to interact with the simulation. However, this approach
    tends to be the fastest.

    .. code-block:: python

        from pyswmm import Simulation

        sim = Simulation('tests/data/model_weir_setting.inp')
        sim.execute()
    """

    def __init__(self,
                 inputfile,
                 reportfile=None,
                 outputfile=None,
                 sim_preconfig=None):
        # sim_config enables a find/replace to be fun on the source input file
        # to create the new INP file.
        if sim_preconfig:
            if not isinstance(sim_preconfig, SimulationPreConfig):
                raise (Exception("Invalid Simulation Preconfig Instance."))
            else:
                if not sim_preconfig.input_file:
                    sim_preconfig.input_file = inputfile
                inputfile = sim_preconfig.apply_changes()

        # Add Simulation State Manager to Prevent Multiple Instances of
        # SWMM to be opened in one instance of Python
        if _sim_state_instance.sim_is_instantiated:
            raise(MultiSimulationError("Multi-Simulation Error."))

        self._model = PySWMM(inputfile, reportfile, outputfile)
        self._model.swmm_open()
        self._is_open = True
        _sim_state_instance.sim_is_instantiated = self._is_open
        self._advance_seconds = None
        self._is_started = False
        self._terminate_request = False
        self._callbacks = {
            "before_start": None,
            "after_start": None,
            "before_step": None,
            "after_step": None,
            "before_end": None,
            "after_end": None,
            "after_close": None
        }

    def __enter__(self):
        """
        Examples:

        .. code-block:: python

            from pyswmm import Simulation

            with Simulation('tests/data/model_weir_setting.inp') as sim:
                for step in sim:
                    print(sim.current_time)

        .. code-block::

            >>> 2015-11-01 14:00:30
            >>> 2015-11-01 14:01:00
            >>> 2015-11-01 14:01:30
            >>> 2015-11-01 14:02:00

        """
        return self

    def __iter__(self):
        """Iterator over Simulation"""
        return self

    def start(self):
        """Start Simulation (no longer suggested to user)."""
        if not self._is_started:
            # Execute Callback Hooks Before Start
            self._execute_callback(self._before_start())
            self._model.swmm_start(True)
            # Execute Callback Hooks After Start
            self._execute_callback(self._after_start())
            self._is_started = True

    def __next__(self):
        """Next"""
        # Start Simulation
        self.start()
        # Check if simulation termination request was made
        if self._terminate_request:
            self._execute_callback(self._before_end())
            raise StopIteration
        # Execute Callback Hooks Before Simulation Step
        self._execute_callback(self._before_step())
        # Simulation Step Amount
        if self._advance_seconds is None:
            time = self._model.swmm_step()
        else:
            time = self._model.swmm_stride(self._advance_seconds)
        # Execute Callback Hooks After Simulation Step
        self._execute_callback(self._after_step())
        if time <= 0.0:
            self._execute_callback(self._before_end())
            raise StopIteration
        return self._model

    def __exit__(self, *a):
        """close"""
        if self._is_started:
            self._model.swmm_end()
            self._is_started = False
            # Execute Callback Hooks After Simulation End
            self._execute_callback(self._after_end())
        if self._is_open:
            self._model.swmm_close()
            self._is_open = False
            # Execute Callback Hooks After Simulation Closes
            self._execute_callback(self._after_close())
        _sim_state_instance.sim_is_instantiated = self._is_open

    @staticmethod
    def _is_callback(callable_object):
        """Checks if arugment is a function/method."""
        if not callable(callable_object):
            error_msg = 'Requires Callable Object, not {}'.format(
                type(callable_object))
            raise (PYSWMMException(error_msg))
        else:
            return True

    def _execute_callback(self, callback):
        """Runs the callback."""
        if callback:
            try:
                callback()
            except PYSWMMException:
                error_msg = "Callback Failed"
                raise PYSWMMException((error_msg))

    @property
    def _isOpen(self) -> bool:
        """._isOpen is set for Deprecation """
        warn('This method will be deprecated in PySWMM-v2.1',
             DeprecationWarning, stacklevel=2)
        return self.sim_is_open

    @property
    def sim_is_open(self) -> bool:
        """Check is Model is Open

        Examples:

        .. code-block:: python

            from pyswmm import Simulation

            with Simulation('tests/data/model_weir_setting.inp') as sim:
                print(sim.sim_is_open)

        .. code-block::

            >>> True
        """
        return self._is_open

    @property
    def _isStarted(self) -> bool:
        """._isSpen is set for Deprecation """
        warn('This method will be deprecated in PySWMM-v2.1',
             DeprecationWarning, stacklevel=2)
        return self.sim_is_started

    @property
    def sim_is_started(self) -> bool:
        """Check is Simulation is Started

        Examples:

        .. code-block:: python

            from pyswmm import Simulation

            with Simulation('tests/data/model_weir_setting.inp') as sim:
                print(sim.sim_is_started)
                for step in sim:
                    print(sim.sim_is_started)

        .. code-block::

            >>> False
            >>> True
        """
        return self._is_started

    def initial_conditions(self, init_conditions):
        """
        Starting in PySWMM-v2 this method/function is set to be
        deprecated.  For setting initial depths refer to the
        Simulation.add_before_start() callback. If the user's goal is to
        set the initial link settings, instead use Simulation.add_after_start().
        """
        warn('This method was deprecated in PySWMM-v2',
             DeprecationWarning, stacklevel=2)

    def step_advance(self, advance_seconds):
        """
        Advances the model by X number of seconds instead of
        intervening at every routing step.  This does not change
        the routing step for the simulation; only lets python take
        back control after each advance period.

        :param int advance_seconds: Seconds to Advance simulation

        Examples:

        .. code-block:: python

            from pyswmm import Simulation

            with Simulation('tests/data/model_weir_setting.inp') as sim:
                sim.step_advance(30)
                for step in sim:
                    print(sim.current_time)

        .. code-block::

            >>> 2015-11-01 14:00:30
            >>> 2015-11-01 14:01:00
            >>> 2015-11-01 14:01:30
            >>> 2015-11-01 14:02:00
        """
        self._advance_seconds = advance_seconds

    def terminate_simulation(self):
        """
        Inserts a request to stop a simulation and cleanly executing the callbacks.

        Examples:

        .. code-block:: python

            with Simulation("model") as sim:
                nodeXYZ = Nodes(sim)["nodeZYX"]

                def before_step_callback():
                    if nodeXYZ.depth > 8:
                        sim.terminate_simulation()

                # Setting Callbacks
                sim.add_before_step(before_step_callback)

                for ind, step in enumerate(sim):
                    # Now simulation will end early if the depth is > 8
                    pass
        """
        self._terminate_request = True

    def report(self):
        """
        Writes to report file after simulation (no longer suggested for user).

        Examples:

        .. code-block:: python

            from pyswmm import Simulation

            with Simulation('tests/data/model_weir_setting.inp') as sim:
                for step in sim:
                    pass
                sim.report()
        """
        self._model.swmm_report()

    def close(self):
        """
        Intialize a simulation and iterate through a simulation. This
        approach requires some clean up. No longer recommended that the user
        call this function directly.
        """
        self.__exit__()

    def execute(self):
        """
        Open an input file, run SWMM, then close the file.

        Examples:

        .. code-block:: python

            sim = Simulation('tests/data/model_weir_setting.inp')
            sim.execute()
        """
        self._model.swmmExec()
        # swmm exec brings the simulation to a close therefore we
        # need to tell the sim state manager that we are free to
        # open another a simulation.
        self._is_open = False
        _sim_state_instance.sim_is_instantiated = self._is_open

    @property
    def engine_version(self):
        """
        Retrieves the SWMM Engine Version.

        :return: Engine Version
        :rtype: LooseVersion

        Examples:

        .. code-block:: python

            with Simulation('tests/data/model_weir_setting.inp') as sim:
                print(sim.engine_version)

        .. code-block::

            >>> 5.1.14
        """
        return self._model.swmm_getVersion()

    @property
    def runoff_error(self):
        """
        Retrieves the Runoff Mass Balance Error.

        :return: Runoff Mass Balance Error
        :rtype: float

        Examples:

        .. code-block:: python

            with Simulation('tests/data/model_weir_setting.inp') as sim:
               sim.execute()
               print(sim.runoff_error)

        .. code-block::

            >>> 0.01
        """
        return self._model.swmm_getMassBalErr()[0]

    @property
    def flow_routing_error(self):
        """
        Retrieves the Flow Routing Mass Balance Error.

        :return: Flow Routing Mass Balance Error
        :rtype: float

        Examples:

        .. code-block:: python

            with Simulation('tests/data/model_weir_setting.inp') as sim:
               sim.execute()
               print(sim.flow_routing_error)

        .. code-block::

            >>> 0.01
        """
        return self._model.swmm_getMassBalErr()[1]

    @property
    def quality_error(self):
        """
        Retrieves the Quality Routing Mass Balance Error.

        :return: Quality Routing Mass Balance Error
        :rtype: float

        Examples:

        .. code-block:: python

            with Simulation('tests/data/model_weir_setting.inp') as sim:
               sim.execute()
               print(sim.quality_error)

        .. code-block::

            >>> 0.01
        """
        return self._model.swmm_getMassBalErr()[2]

    @property
    def start_time(self):
        """Get/set Simulation start time.

        Examples:

        .. code-block:: python

            from pyswmm import Simulation

            with Simulation('tests/data/model_weir_setting.inp') as sim:
                print(sim.start_time)
                sim.start_time = datetime(2015,5,10,15,15,1)

        .. code-block::

            >>> datetime.datetime(2015,5,10,15,15,1)
        """
        return self._model.getSimulationDateTime(
            SimulationTime.StartDateTime.value)

    @start_time.setter
    def start_time(self, dtimeval):
        """Set simulation Start time"""
        self._model.setSimulationDateTime(SimulationTime.StartDateTime.value,
                                          dtimeval)

    @property
    def end_time(self):
        """Get/set Simulation end time.

        Examples:

        .. code-block:: python

            from pyswmm import Simulation

            with Simulation('tests/data/model_weir_setting.inp') as sim:
                print(sim.end_time)
                sim.end_time = datetime(2016,5,10,15,15,1)

        .. code-block::

            >>> datetime.datetime(2016,5,10,15,15,1)
        """
        return self._model.getSimulationDateTime(
            SimulationTime.EndDateTime.value)

    @end_time.setter
    def end_time(self, dtimeval):
        """Set simulation End time."""
        self._model.setSimulationDateTime(SimulationTime.EndDateTime.value,
                                          dtimeval)

    @property
    def report_start(self):
        """Get/set Simulation report start time.

        Examples:

        .. code-block:: python

            from pyswmm import Simulation

            with Simulation('tests/data/model_weir_setting.inp') as sim:
                print(sim.report_start)
                sim.report_start = datetime(2015,5,10,15,15,1)

        .. code-block::

            >>> datetime.datetime(2015,5,10,15,15,1)
        """
        return self._model.getSimulationDateTime(
            SimulationTime.ReportStart.value)

    @report_start.setter
    def report_start(self, dtimeval):
        """Set simulation report start time."""
        self._model.setSimulationDateTime(SimulationTime.ReportStart.value,
                                          dtimeval)

    @property
    def flow_units(self):
        """
        Get Simulation Units (CFS, GPM, MGD, CMS, LPS, MLD).

        :return: Flow Unit
        :rtype: str

        Examples:

        .. code-block:: python

            from pyswmm import Simulation

            with Simulation('tests/data/model_weir_setting.inp') as sim:
                print(sim.flow_units)

        .. code-block::

            >>> CFS
        """
        return self._model.getSimUnit(SimulationUnits.FlowUnits.value)

    @property
    def system_units(self):
        """Get system units (US, SI).

        :return: System Unit
        :rtype: str

        Examples:

        .. code-block:: python

            from pyswmm import Simulation

            with Simulation('tests/data/model_weir_setting.inp') as sim:
                print(sim.system_units)

        .. code-block::

            >>> US
        """
        return self._model.getSimUnit(SimulationUnits.UnitSystem.value)

    @property
    def current_time(self):
        """Get Simulation Current Time.

        :return: Current Simulation Time
        :rtype: Datetime

        Examples:

        .. code-block:: python

            from pyswmm import Simulation

            with Simulation('tests/data/model_weir_setting.inp') as sim:
                for step in sim:
                    print(sim.current_time)

        .. code-block::

            >>> 2015-11-01 14:00:30
            >>> 2015-11-01 14:01:00
            >>> 2015-11-01 14:01:30
            >>> 2015-11-01 14:02:00
        """
        return self._model.getCurrentSimulationTime()

    @property
    def percent_complete(self):
        """Get Simulation Percent Complete.

        :return: Current Percent Complete
        :rtype: Datetime

        Examples:

        .. code-block:: python

            from pyswmm import Simulation

            with Simulation('tests/data/model_weir_setting.inp') as sim:
                for step in sim:

        .. code-block::

            >>> 0.01
            >>> 0.25
            >>> 0.50
            >>> 0.75
        """
        dt = self.current_time - self.start_time
        total_time = self.end_time - self.start_time
        return float(dt.total_seconds()) / total_time.total_seconds()

    def use_hotstart(self, hotstart_file):
        """
        Use a hotstart file to initialize the simulation.

        This must be run before the simualation loop but within
        the simulation context manager.

        :param str hotstart_file: Path to hotstart file.

        .. code-block:: python

            with Simulation('model_weir_setting.inp') as sim:
                sim.use_hotstart("path_to_hotstart.hsf")

                for ind, step in enumerate(sim):
                    break

        """
        self._model.swmm_use_hotstart(hotstart_file)

    def save_hotstart(self, hotstart_file):
        """
        Save the current state of the model to a hotstart file.

        This can be run at any point during the simultion.

        :param str hotstart_file: Path to hotstart file.

        .. code-block:: python

            with Simulation('model_weir_setting.inp') as sim:
                for ind, step in enumerate(sim):
                    if ind == 10:
                        sim.save_hotstart('new_hsf.HSF')

        """
        self._model.swmm_save_hotstart(hotstart_file)

    def _before_start(self):
        """Get Before Start Callback.

        :return: Callbacks
        """
        return self._callbacks["before_start"]

    def add_before_start(self, callback):
        """
        Add callback function/method/object to execute before
        the simlation starts. Needs to be callable.

        :param func callback: Callable Object

        (See self.add_after_close() for more details)
        """
        if self._is_callback(callback):
            self._callbacks["before_start"] = callback

    def _after_start(self):
        """Get After Start Callback.

        :return: Callbacks
        """
        return self._callbacks["after_start"]

    def add_after_start(self, callback):
        """
        Add callback function/method/object to execute after
        a simlation start. Needs to be callable.  This callback allows
        setting initial link target_settings (such as an orifice).

        :param func callback: Callable Object

        (See self.add_after_close() for more details)
        """
        if self._is_callback(callback):
            self._callbacks["after_start"] = callback

    def _before_step(self):
        """Get Before Step Callback.

        :return: Callbacks
        """
        return self._callbacks["before_step"]

    def add_before_step(self, callback):
        """
        Add callback function/method/object to execute before
        a simlation step. Needs to be callable.

        :param func callback: Callable Object

        (See self.add_after_close() for more details)
        """
        if self._is_callback(callback):
            self._callbacks["before_step"] = callback

    def _after_step(self):
        """Get After Step Callback.

        :return: Callbacks
        """
        return self._callbacks["after_step"]

    def add_after_step(self, callback):
        """
        Add callback function/method/object to execute after
        a simlation step. Needs to be callable.

        :param func callback: Callable Object

        (See self.add_after_close() for more details)
        """
        if self._is_callback(callback):
            self._callbacks["after_step"] = callback

    def _before_end(self):
        """Get Before End Callback.

        :return: Callbacks
        """
        return self._callbacks["before_end"]

    def add_before_end(self, callback):
        """
        Add callback function/method/object to execute after
        the simulation ends. Needs to be callable.

        :param func callback: Callable Object

        (See self.add_after_close() for more details)
        """
        if self._is_callback(callback):
            self._callbacks["before_end"] = callback

    def _after_end(self):
        """Get After End Callback.

        :return: Callbacks
        """
        return self._callbacks["after_end"]

    def add_after_end(self, callback):
        """
        Add callback function/method/object to execute after
        the simulation ends. Needs to be callable.

        :param func callback: Callable Object

        (See self.add_after_close() for more details)
        """
        if self._is_callback(callback):
            self._callbacks["after_end"] = callback

    def _after_close(self):
        """Get After Close Callback.

        :return: Callbacks
        """
        return self._callbacks["after_close"]

    def add_after_close(self, callback):
        """
        Add callback function/method/object to execute after
        the simulation is closed. Needs to be callable.

        :param func callback: Callable Object

        .. code-block:: python

            from pyswmm import Simulation

            def test_callback():
                print("CALLBACK - Executed")

            with Simulation('tests/data/model_weir_setting.inp') as sim:

                sim.before_start(test_callback) #<- pass function handle.
                print("Waiting to Start")
                for ind, step in enumerate(sim):
                    print("Step {}".format(ind))
                print("Complete!")
            print("Closed")

        .. code-block::

            >>> "Waiting to Start"
            >>> "CALLBACK - Executed"
            >>> "Step 0"
            >>> "Step 1"
            >>> "Complete!"
            >>> "Closed"
        """
        if self._is_callback(callback):
            self._callbacks["after_close"] = callback


class SimulationPreConfig():
    """
    This class was developed to introduce a simple way to programmatically
    adjust nearly all model parameters. Once the user instantiates the
    `SimulationPreConfig` object the method `add_update_by_token` can be called
    for each parameter (by index) that they would like to update. The parameter
    limits are still up to the user to get right as per the SWMM user's guide.

    The arguments are as follows.  In the Base INP file:

    .. code-block:: python

        [SUBCATCHMENTS]
        ;;                                         Total   Pcnt. Pcnt.  Curb
        ;;Name   Raingage         Outlet   Area    Imperv  Width Slope  Length
        ;;------ ---------------- -------- ------- ------- ----- ------ --------
        S1       SCS_24h_Type_I_1in J1     1       100     500   0.5    0


    .. code-block:: python

        from pyswmm import Simulation, SimulationPreConfig, Subcatchments

        # Create Config Handle
        sim_conf = SimulationPreConfig()

        # Specifying the update parameters
        # Parameter Order:
        # Section, Object ID, Parameter Index, New Value, Obj Row Num (optional)
        sim_conf.add_update_by_token("SUBCATCHMENTS", "S1", 2, "J2")
        sim_conf.add_update_by_token("TIMESERIES", "SCS_24h_Type_I_1in", 2, 2, 5)

        with Simulation(<path-to-inp>, sim_preconfig = sim_conf) as sim:
            S1 = Subcatchments(sim)["S1"]
            print(S1.connection)

            for step in sim:
                pass

    .. code-block:: python

        >>> (2, 'J2')

    In the New INP file:

    .. code-block:: python

        [SUBCATCHMENTS]
        ;;                                         Total   Pcnt. Pcnt.  Curb
        ;;Name   Raingage         Outlet   Area    Imperv  Width Slope  Length
        ;;------ ---------------- -------- ------- ------- ----- ------ --------
        S1       SCS_24h_Type_I_1in J2     1       100     500   0.5    0

    """

    def __init__(self):
        self._filename_suffix = "_mod"
        self._modifications = {}
        self._source_input_name = None

    @property
    def input_file(self):
        """
        This is set by the `Simulation` class but can also be set directly
        if the user wants to simply use this class for find replace.

        Examples:

        .. code-block:: python

            sim_conf = SimulationPreConfig()
            sim_conf.input_file = "./model_weir_setting.inp"

        .. code-block::

            >>> datetime.datetime(2016,5,10,15,15,1)
        """
        self._source_input_name

    @input_file.setter
    def input_file(self, inp_path):
        """"""
        self._source_input_name = inp_path

    @property
    def filename_suffix(self):
        """
        If the user wants to modify the new file name simply use this class
        for find replace.

        Examples:

        .. code-block:: python

            sim_conf = SimulationPreConfig()
            sim_conf.filename_suffix = "_a"

        .. code-block::

            >>> datetime.datetime(2016,5,10,15,15,1)
        """
        return self._filename_suffix

    @filename_suffix.setter
    def filename_suffix(self, suffix: str):
        """"""
        self._filename_suffix = suffix

    def add_update_by_token(self, section: str, obj_id: str,
                            index: int, new_val, row_num=0):
        """
        This method allows the user to give the parmeter to be updated and
        where this value should be set in the input file.

        :param str section: Section name (such as "JUNCTIONS")
        :param str obj_id: the SWMM object ID name (such as "J1")
        :param int index: The index of the parameter in the row to update (0 is first index)
        :param new_val: The new value (can be any normal data type)
        :param ind row_num: If multiple rows exist for an object like "HYDROGRAPHS" (0 is first index)

        .. code-block:: python

            from pyswmm import SimulationPreConfig

            # Create Config Handle
            sim_conf = SimulationPreConfig()

            # Specifying the update parameters
            # Parameter Order:
            # New Value, Section, Object ID, Parameter Index, Obj Row Num (optional)
            sim_conf.add_update_by_token("SUBCATCHMENTS", "S1", 2, "J2")
            sim_conf.add_update_by_token("TIMESERIES", "SCS_24h_Type_I_1in", 2, 2, 5)

        """
        section = section.lower()
        obj_id = obj_id.lower()

        if section not in self._modifications.keys():
            self._modifications[section] = {}
        if obj_id not in self._modifications[section].keys():
            self._modifications[section][obj_id] = {}
        if row_num not in self._modifications[section][obj_id.lower()].keys():
            self._modifications[section][obj_id][row_num] = {}

        self._modifications[section][obj_id][row_num][index] = new_val

    def apply_changes(self):
        """
        If the user wants to modify the new file name simply use this class
        for find replace then Apply those changes.

        Examples:

        .. code-block:: python

            sim_conf = SimulationPreConfig()
            sim_conf.input_file = "./model_weir_setting.inp"
            sim_conf.filename_suffix = "_a"

            sim_conf.apply_changes()

        .. code-block::

            >>> datetime.datetime(2016,5,10,15,15,1)

        """
        mods = self._modifications
        if not self._source_input_name:
            raise (Exception("No Source INP set"))

        def write_line(fl_handle, ln):
            end = ''
            if not ln.endswith("\n"):
                end = '\n'
            fl_handle.write(ln+end)

        dest_file = self._source_input_name[:-4] \
            + self._filename_suffix + '.inp'

        fl_source = open(self._source_input_name, 'r')
        fl_destin = open(dest_file, 'w')

        section_replacements = False
        section = None
        id_ref = None
        row_count = 0

        for ln in fl_source:
            if '[' in ln and ']' in ln and ln.strip()[0]!=';':
                ln_orig = ln
                ln = ln.strip()
                ln = ln.replace("[", '')
                ln = ln.replace("]", '')
                section = ln.lower()
                section_replacements = False
                # Only modify sections if there are edits.
                if section in mods.keys():
                    section_replacements = True
                id_ref = None
                row_count = 0
                write_line(fl_destin, ln_orig)
            elif ln.startswith(';') or len(ln.split()) == 0:
                write_line(fl_destin, ln)
            elif not section_replacements:
                write_line(fl_destin, ln)
            else:
                ln = ln.strip()
                ln_split = ln.split()
                if id_ref == ln_split[0].lower():
                    row_count += 1
                else:
                    id_ref = ln_split[0].lower()
                    row_count = 0

                ln_mod = ln_split
                if id_ref in mods[section].keys():
                    if row_count in mods[section][id_ref].keys():
                        for index in mods[section][id_ref][row_count].keys():
                            if index >= 0 and index < len(ln_split):
                                ln_mod[index] = mods[section][id_ref][row_count][index]
                                ln_out = "     ".join([str(v) for v in ln_mod])
                            else:
                                raise (Exception("{0} {1} {2} index {3} out of bounds".format(
                                    section, id_ref,
                                    row_count, index)))
                    else:
                        ln_out = ln
                else:
                    ln_out = ln
                write_line(fl_destin, ln_out)
        fl_source.close()
        fl_destin.close()

        return dest_file



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014 Bryant E. McDonnell
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------

# Local imports
from pyswmm import Links, Nodes, Simulation
# from pyswmm.swmm5 import PySWMM
from pyswmm.tests.data import (MODEL_PUMP_SETTINGS_PATH, MODEL_STORAGE_PUMP,
                               MODEL_WEIR_SETTING_PATH)


def test_links_1():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        print("\n\n\nLINKS\n")
        c1c2 = Links(sim)["C1:C2"]
        assert (c1c2.linkid == "C1:C2")
        assert (c1c2.is_conduit() == True)
        assert (c1c2.is_pump() == False)
        assert (c1c2.is_orifice() == False)
        assert (c1c2.is_weir() == False)
        assert (c1c2.is_outlet() == False)
        assert (c1c2.connections == ("J1", "J2"))
        assert (c1c2.inlet_node == "J1")
        assert (c1c2.outlet_node == "J2")


def test_links_2():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        link_names = ["C1", "C1:C2", "C2", "C3"]
        assert "DUMMY_LINK" not in Links(sim)
        for link in Links(sim):
            assert (link.linkid in link_names)
            link.flow_limit = 10
            assert (link.flow_limit == 10)


def test_links_3():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        print("\n\n\nLINKS\n")
        c1c2 = Links(sim)["C1:C2"]
        sim.step_advance(300)
        for ind, step in enumerate(sim):
            if ind > 15:
                c1c2.flow_limit = 1
            if ind > 30:
                c1c2.flow_limit = 0

            if ind > 16 and ind <= 30:
                assert (c1c2.flow == 1)


def test_links_4():
    with Simulation(MODEL_PUMP_SETTINGS_PATH) as sim:
        peak_pump_rate = 20  # cfs
        print("\n\n\nLINKS\n")
        c3 = Links(sim)["C3"]

        sim.step_advance(300)
        for ind, step in enumerate(sim):
            # setting adjustment
            if ind == 15:
                c3.target_setting = 0.9
            if ind == 20:
                c3.target_setting = 0.8
            if ind == 25:
                c3.target_setting = 0.7
            if ind == 30:
                c3.target_setting = 0.6
            if ind == 35:
                c3.target_setting = 2.0
            if ind == 40:
                c3.target_setting = 0.4
            if ind == 45:
                c3.target_setting = 0.3
            if ind == 50:
                c3.target_setting = 0.2
            if ind == 55:
                c3.target_setting = 0.1
            if ind == 60:
                c3.target_setting = 0.0
            if ind == 65:
                c3.target_setting = 1.0

            # Check Results
            if ind == 16:
                assert (c3.target_setting == 0.9)
                assert (c3.flow == 0.9 * peak_pump_rate)
            if ind == 21:
                assert (c3.target_setting == 0.8)
                assert (c3.flow == 0.8 * peak_pump_rate)
            if ind == 26:
                assert (c3.target_setting == 0.7)
                assert (c3.flow == 0.7 * peak_pump_rate)
            if ind == 31:
                assert (c3.target_setting == 0.6)
                assert (c3.flow == 0.6 * peak_pump_rate)
            if ind == 36:
                assert (c3.target_setting == 2.0)
                assert (c3.flow >= peak_pump_rate)
            if ind == 41:
                assert (c3.target_setting == 0.4)
                assert (c3.flow == 0.4 * peak_pump_rate)
            if ind == 46:
                assert (c3.target_setting == 0.3)
                assert (c3.flow == 0.3 * peak_pump_rate)
            if ind == 51:
                assert (c3.target_setting == 0.2)
                assert (c3.flow == 0.2 * peak_pump_rate)
            if ind == 56:
                assert (c3.target_setting == 0.1)
                assert (c3.flow == 0.1 * peak_pump_rate)
            if ind == 61:
                assert (c3.target_setting == 0.0)
                assert (c3.flow == 0.0 * peak_pump_rate)
            if ind == 66:
                assert (c3.target_setting == 1.0)
                assert (c3.flow == 1.0 * peak_pump_rate)


def test_links_5():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        weir_pump_rate = 5  # cfs
        print("\n\n\nWEIR\n")
        c3 = Links(sim)["C3"]

        sim.step_advance(300)
        for ind, step in enumerate(sim):

            # setting adjustment
            if ind == 15:
                c3.target_setting = 0.9
            if ind == 20:
                c3.target_setting = 0.8
            if ind == 25:
                c3.target_setting = 0.7
            if ind == 30:
                c3.target_setting = 0.6
            if ind == 35:
                c3.target_setting = 0.5
            if ind == 40:
                c3.target_setting = 0.4
            if ind == 45:
                c3.target_setting = 0.3
            if ind == 50:
                c3.target_setting = 0.2
            if ind == 55:
                c3.target_setting = 0.1
            if ind == 60:
                c3.target_setting = 0.0
            if ind == 65:
                c3.target_setting = 1.0

            # Check Results
            if ind == 16:
                assert (c3.target_setting == 0.9)
                assert (c3.flow >= 0.9 * weir_pump_rate)
            if ind == 21:
                assert (c3.target_setting == 0.8)
                assert (c3.flow >= 0.8 * weir_pump_rate)
            if ind == 26:
                assert (c3.target_setting == 0.7)
                assert (c3.flow >= 0.7 * weir_pump_rate)
            if ind == 31:
                assert (c3.target_setting == 0.6)
                assert (c3.flow >= 0.6 * weir_pump_rate)
            if ind == 36:
                assert (c3.target_setting == 0.5)
                assert (c3.flow >= 0.5 * weir_pump_rate)
            if ind == 41:
                assert (c3.target_setting == 0.4)
                assert (c3.flow >= 0.4 * weir_pump_rate)
            if ind == 46:
                assert (c3.target_setting == 0.3)
                assert (c3.flow >= 0.3 * weir_pump_rate)
            if ind == 51:
                assert (c3.target_setting == 0.2)
                assert (c3.flow >= 0.15 * weir_pump_rate)
            if ind == 56:
                assert (c3.target_setting == 0.1)
                assert (c3.flow >= 0.05 * weir_pump_rate)
            if ind == 61:
                assert (c3.target_setting == 0.0)
                assert (c3.flow == 0.0 * weir_pump_rate)
            if ind == 66:
                assert (c3.target_setting == 1.0)
                assert (c3.flow >= 1.0 * weir_pump_rate)


def test_links_6():
    with Simulation(MODEL_STORAGE_PUMP) as sim:
        print("\n\n\nConduits\n")
        link = Links(sim)["C2"]

        sim.step_advance(300)
        for ind, step in enumerate(sim):
            if ind % 1000 == 0:
                print(link.conduit_statistics)


def test_links_7():
    with Simulation(MODEL_STORAGE_PUMP) as sim:
        print("\n\n\nPUMPS\n")
        link = Links(sim)["P1"]

        sim.step_advance(300)
        for ind, step in enumerate(sim):
            if ind % 1000 == 0:
                print(link.pump_statistics)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014 Bryant E. McDonnell
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
"""System module for the pythonic interface to SWMM5."""

# Local imports
from pyswmm.swmm5 import PYSWMMException


class SystemStats(object):
    """
    System-Wide Flow and Runoff Routing Accumulation Volume Methods.

    :param object model: Open Model Instance

    Examples:

    >>> from pyswmm import Simulation, SystemStats
    >>>
    >>> with Simulation('tests/data/TestModel1_weirSetting.inp') as sim:
    ...     system_routing = SystemStats(sim)
    ...
    ...     for step in sim:
    ...         print(system_routing.routing_stats)
    ...         print(system_routing.runoff_stats)
    """

    def __init__(self, model):
        if not model._model.fileLoaded:
            raise PYSWMMException("SWMM Model Not Open")
        self._model = model._model

    # --- Get Parameters
    # -------------------------------------------------------------------------
    @property
    def routing_stats(self):
        """
        Get rolling/cumulative routing stats. Follow Data are returned:

        +--------------------+
        | dry_weather_inflow |
        +--------------------+
        | wet_weather_inflow |
        +--------------------+
        | groundwater_inflow |
        +--------------------+
        | II_inflow          |
        +--------------------+
        | external_inflow    |
        +--------------------+
        | flooding           |
        +--------------------+
        | outflow            |
        +--------------------+
        | evaporation_loss   |
        +--------------------+
        | seepage_loss       |
        +--------------------+
        | reacted            |
        +--------------------+
        | initial_storage    |
        +--------------------+
        | final_storage      |
        +--------------------+
        | routing_error      |
        +--------------------+

        :return: Statistics
        :rtype: dict
        """
        return self._model.flow_routing_stats()

    @property
    def runoff_stats(self):
        """
        Get rolling/cumulative runoff stats. Follow Data are returned:

        +------------------+
        | rainfall         |
        +------------------+
        | evaporation      |
        +------------------+
        | infiltration     |
        +------------------+
        | runoff           |
        +------------------+
        | drains           |
        +------------------+
        | runon            |
        +------------------+
        | init_storage     |
        +------------------+
        | final_storage    |
        +------------------+
        | init_snow_cover  |
        +------------------+
        | final_snow_cover |
        +------------------+
        | snow_removed     |
        +------------------+
        | routing_error    |
        +------------------+

        :return: Statistics
        :rtype: dict
        """
        return self._model.runoff_routing_stats()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2019 Bryant E. McDonnell
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
"""Raingages module for the pythonic interface to SWMM5."""

# Local imports
from pyswmm.swmm5 import PYSWMMException
from pyswmm.toolkitapi import RainGageResults, ObjectType


class RainGages(object):
    """
    Rain Gages Iterator Methods.

    :param object model: Open Model Instance

    Examples:

    >>> from pyswmm import Simulation, Nodes
    >>>
    >>> with Simulation('tests/data/TestModel1_weirSetting.inp') as sim:
    ...     for raingage in RainGages(sim):
    ...         print(raingage)
    ...         print(raingage.raingageid)
    ...
    >>> <swmm5.RainGage object at 0x031B0350>
    >>> Gage1
    >>> <swmm5.RainGage object at 0x030693D0>
    >>> Gage4
    >>> <swmm5.RainGage object at 0x031B0350>
    >>> Gage3
    >>> <swmm5.RainGage object at 0x030693D0>
    >>> Gage10

    Iterating over Nodes Object

    >>> raingages = RainGages(sim)
    >>> for raingage in raingages:
    ...     print(raingage.raingageid)
    >>> Gage1
    >>> Gage4
    >>> Gage3
    >>> Gage10

    Testing Existence

    >>> raingages = RainGages(sim)
    >>> "Gage1" in raingages
    >>> True

    Initializing a node Object

    >>> raingages = RainGages(sim)
    >>> gage1 = raingages['Gage1']
    >>> print(gage1.total_precip)
    >>> 0.04
    >>>
    >>> gage1.total_precip = 1
    >>> print(gage1.total_precip)
    >>> 1
    """

    def __init__(self, model):
        if not model._model.fileLoaded:
            raise PYSWMMException("SWMM Model Not Open")
        self._model = model._model
        self._cuindex = 0
        self._nRaingages = self._model.getProjectSize(ObjectType.GAGE.value)

    def __len__(self):
        """
        Return number of raingages. Use the expression 'len(RainGages(sim))'.

        :return: Number of Raingages
        :rtype: int

        """
        return self._model.getProjectSize(ObjectType.GAGE.value)

    def __contains__(self, raingageid):
        """
        Checks if Rain Gage ID exists.

        :return: ID Exists
        :rtype: bool
        """
        return self._model.ObjectIDexist(ObjectType.GAGE.value, raingageid)

    def __getitem__(self, raingageid):
        if self.__contains__(raingageid):
            rg = RainGage(self._model, raingageid)
            return rg
        else:
            raise PYSWMMException(
                "Raingage ID: {} Does not Exist".format(raingageid))

    def __iter__(self):
        return self

    def __next__(self):
        if self._cuindex < self._nRaingages:
            raingageobject = self.__getitem__(self._raingageid)
            self._cuindex += 1  # Next Iteration
            return raingageobject
        else:
            raise StopIteration()

    @property
    def _raingageid(self):
        """Node ID."""
        return self._model.getObjectId(ObjectType.GAGE.value, self._cuindex)


class RainGage(object):
    """
    Raingage Methods.

    :param object model: Open Model Instance
    :param str raingageid: Raingage ID

    Examples:

    >>> from pyswmm import Simulation, Raingages
    >>>
    >>> with Simulation('tests/data/TestModel1_weirSetting.inp') as sim:
    ...     rg1 = Raingages(sim)["Gage1"]
    ...     print(rg1.raingageid)
    ...     for step in sim:
    ...         print(rg1.total_precip)
    ... Gage1
    ... 0
    ... 10
    """

    def __init__(self, model, raingageid):
        if not model.fileLoaded:
            raise PYSWMMException("SWMM Model Not Open")
        if raingageid not in model.getObjectIDList(ObjectType.GAGE.value):
            raise PYSWMMException("ID Not valid")
        self._model = model
        self._raingageid = raingageid

    # --- Get Parameters
    # -------------------------------------------------------------------------
    @property
    def raingageid(self):
        """
        Get Rain Gage ID.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, RainGages
        >>>
        >>> with Simulation('tests/data/TestModel1_weirSetting.inp') as sim:
        ...     rg = RainGage(sim)["Gage1"]
        ...     print(rg.raingageid)
        >>> Gage1
        """
        return self._raingageid

    @property
    def total_precip(self):
        """
        Get/set raingage total precipitation rate (like in/hr or mm/hr).

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, RainGages
        >>>
        >>> with Simulation('tests/data/TestModel1_weirSetting.inp') as sim:
        ...     rg1 = RainGages(sim)["Gage1"]
        ...     print(rg1.total_precip)
        >>> 1.0

        Setting the value

        >>> from pyswmm import Simulation, RainGages
        >>>
        >>> with Simulation('tests/data/TestModel1_weirSetting.inp') as sim:
        ...     rg1 = RainGages(sim)["Gage1"]
        ...     print(rg1.total_precip)
        ...     rg1.total_precip = 0.2
        ...     print(rg1.total_precip)
        >>> 1.0
        >>> 0.2
        """
        return self._model.getGagePrecip(self._raingageid, RainGageResults.total_precip)

    @total_precip.setter
    def total_precip(self, param):
        """Set Total Precipitation Rate (in/hr or mm/hr)."""
        self._model.setGagePrecip(self._raingageid, param)

    @property
    def rainfall(self):
        """
        Get raingage total rainfall rate (like in/hr or mm/hr).

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, RainGages
        >>>
        >>> with Simulation('tests/data/TestModel1_weirSetting.inp') as sim:
        ...     rg1 = RainGages(sim)["Gage1"]
        ...     print(rg1.rainfall)
        >>> 1.0
        """
        return self._model.getGagePrecip(self._raingageid, RainGageResults.rainfall)

    @property
    def snowfall(self):
        """
        Get raingage total snowfall rate (like in/hr or mm/hr).

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, RainGages
        >>>
        >>> with Simulation('tests/data/TestModel1_weirSetting.inp') as sim:
        ...     rg1 = RainGages(sim)["Gage1"]
        ...     print(rg1.snowfall)
        >>> 0.0
        """
        return self._model.getGagePrecip(self._raingageid, RainGageResults.snowfall)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2018 Jennifer Wu
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
from pyswmm.toolkitapi import LidLayers, LidResults


def _flux_rate(model, subcatchment, lid_index, layer):
    """
    Get lid net inflow - outflow from previous time step for each lid layer
    ONLY FOR for surface, soil, storage, pave
    :param int layerIndex: layer type (toolkitapi.LidLayers member variable)
    :return: Parameter Value
    :rtype: double
    """
    return model.getLidUFluxRates(subcatchment,
                                  lid_index,
                                  layer)


class Surface(object):
    def __init__(self, model, lidunit):
        self._model = model
        self._lidunit = lidunit
        self._subcatchmentid = lidunit._subcatchmentid
        self._lidid = lidunit._lidid

    @property
    def depth(self):
        """
        Get lid depth of ponded water on surface layer

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.surfDepth.value)

    @property
    def inflow(self):
        """
        Get lid precip. + runon to LID unit

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.surfInflow.value)

    @property
    def infiltration(self):
        """
        Get lid infiltration rate from surface layer

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.surfInfil.value)

    @property
    def evaporation(self):
        """
        Get lid evaporation rate from surface layer

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.surfEvap.value)

    @property
    def outflow(self):
        """
        Get lid outflow from surface layer

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.surfOutflow.value)

    @property
    def flux_rate(self):
        """
        Get lid flux rate from surface layer

        :return: Parameter Value
        :rtype: double
        """
        return _flux_rate(self._model,
                          self._subcatchmentid,
                          self._lidid,
                          LidLayers.surface.value)


class Pavement(object):
    def __init__(self, model, lidunit):
        self._model = model
        self._lidunit = lidunit
        self._subcatchmentid = lidunit._subcatchmentid
        self._lidid = lidunit._lidid

    @property
    def depth(self):
        """
        Get lid depth of water in poroous pavement layer

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.paveDepth.value)

    @property
    def evaporation(self):
        """
        Get lid evaporation from pavement layer

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.paveEvap.value)

    @property
    def percolation(self):
        """
        Get lid percolation from pavement layer

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.pavePerc.value)

    @property
    def flux_rate(self):
        """
        Get lid flux rate from pavement layer

        :return: Parameter Value
        :rtype: double
        """
        return _flux_rate(self._model,
                          self._subcatchmentid,
                          self._lidid,
                          LidLayers.pavement.value)


class Storage(object):
    def __init__(self, model, lidunit):
        self._model = model
        self._lidunit = lidunit
        self._subcatchmentid = lidunit._subcatchmentid
        self._lidid = lidunit._lidid

    @property
    def depth(self):
        """
        Get lid depth of water in storage layer

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.storDepth.value)

    @property
    def inflow(self):
        """
        Get lid inflow rate to storage rate

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.storInflow.value)

    @property
    def exfiltration(self):
        """
        Get lid exfiltration rate from storage layer

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.storExfil.value)

    @property
    def evaporation(self):
        """
        Get lid evaporation rate from storage layer

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.storEvap.value)

    @property
    def drain(self):
        """
        Get lid drain rate from storage layer

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.storDrain.value)

    @property
    def flux_rate(self):
        """
        Get lid flux rate from storage layer

        :return: Parameter Value
        :rtype: double
        """
        return _flux_rate(self._model,
                          self._subcatchmentid,
                          self._lidid,
                          LidLayers.storage.value)


class Soil(object):
    def __init__(self, model, lidunit):
        self._model = model
        self._lidunit = lidunit
        self._subcatchmentid = lidunit._subcatchmentid
        self._lidid = lidunit._lidid

    @property
    def moisture(self):
        """
        Get lid moisture content of biocell soil layer

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.soilMoist.value)

    @property
    def evaporation(self):
        """
        Get lid evaporation from soil layer

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.soilEvap.value)

    @property
    def percolation(self):
        """
        Get lid percolation from soil layer

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.soilPerc.value)

    @property
    def flux_rate(self):
        """
        Get lid flux rate from soil layer

        :return: Parameter Value
        :rtype: double
        """
        return _flux_rate(self._model,
                          self._subcatchmentid,
                          self._lidid,
                          LidLayers.soil.value)


class WaterBalance(object):
    def __init__(self, model, lidunit):
        self._model = model
        self._lidunit = lidunit
        self._subcatchmentid = lidunit._subcatchmentid
        self._lidid = lidunit._lidid

    @property
    def inflow(self):
        """
        Get lid water balance total inflow

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.inflow.value)

    @property
    def evaporation(self):
        """
        Get lid water balance total evaporation

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.evap.value)

    @property
    def infiltration(self):
        """
        Get lid water balance total infiltration

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.infil.value)

    @property
    def surface_flow(self):
        """
        Get lid water balance total surface runoff

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.surfFlow.value)

    @property
    def drain_flow(self):
        """
        Get lid water balance total underdrain flow

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.drainFlow.value)

    @property
    def initial_volume(self):
        """
        Get lid water balance initial stored volume

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.initVol.value)

    @property
    def final_volume(self):
        """
        Get lid water balance final stored volume

        :return: Parameter Value
        :rtype: double
        """
        return self._model.getLidUResult(self._subcatchmentid,
                                         self._lidid,
                                         LidResults.finalVol.value)



class IncompleteSimulation(Exception):
    """Exception raised for incomplete simulation

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class OutputException(Exception):
    """Exception raised for SWMM Output object

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

_multi_sim_message = \
"""\tDue to architectural limitations inside of EPA-SWMM
\tmultiple Simulations cannot be completed within a single
\tinstance of Python.  Consider launching the model runs
\tusing a Python Subprocess and a Queue manager.\n"""

class MultiSimulationError(Exception):
    """Exception raised if more than one SWMM Simulation
    class is trying to be created within a single instance
    of Python.
    """

    def __init__(self, message):

        self.message = message + "\n\n" + _multi_sim_message
        super().__init__(self.message)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014 Bryant E. McDonnell
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
"""Nodes module for the pythonic interface to SWMM5."""
from swmm.toolkit import shared_enum

# Local imports
from pyswmm.swmm5 import PYSWMMException
from pyswmm.toolkitapi import NodeParams, NodeResults, NodePollut, NodeType, ObjectType


class Nodes(object):
    """
    Node Iterator Methods.

    :param object model: Open Model Instance

    Examples:

    >>> from pyswmm import Simulation, Nodes
    >>>
    >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
    ...     for node in Nodes(sim):
    ...         print(node)
    ...         print(node.nodeid)
    <swmm5.Node object at 0x031B0350>
    J1
    <swmm5.Node object at 0x030693D0>
    J2
    <swmm5.Node object at 0x031B0350>
    J3
    <swmm5.Node object at 0x030693D0>
    J0

    Iterating over Nodes Object

    >>> nodes = Nodes(sim)
    >>> for node in nodes:
    ...     print(node.nodeid)
    J0
    J1
    J2
    J3

    Testing Existence

    >>> nodes = Nodes(sim)
    >>> "J1" in nodes
    True

    Initializing a node Object

    >>> nodes = Nodes(sim)
    >>> j1 = nodes['J1']
    >>> print(j1.invert_elevation)
    12
    >>>
    >>> j1.invert_elevation = 200
    >>> print(j1.invert_elevation)
    200
    """

    def __init__(self, model):
        if not model._model.fileLoaded:
            raise PYSWMMException("SWMM Model Not Open")
        self._model = model._model
        self._cuindex = 0
        self._nNodes = self._model.getProjectSize(ObjectType.NODE.value)

    def __len__(self):
        """
        Return number of nodes. Use the expression 'len(Nodes)'.

        :return: Number of Nodes
        :rtype: int

        """
        return self._model.getProjectSize(ObjectType.NODE.value)

    def __contains__(self, nodeid):
        """
        Checks if Node ID exists.

        :return: ID Exists
        :rtype: bool
        """
        return self._model.ObjectIDexist(ObjectType.NODE.value, nodeid)

    def __getitem__(self, nodeid):
        if self.__contains__(nodeid):
            nd = Node(self._model, nodeid)
            _nd = nd
            if nd.is_outfall():
                _nd.__class__ = Outfall
            elif nd.is_storage():
                _nd.__class__ = Storage
            return _nd

        else:
            raise PYSWMMException("Node ID: {} Does not Exist".format(nodeid))

    def __iter__(self):
        return self

    def __next__(self):
        if self._cuindex < self._nNodes:
            nodeobject = self.__getitem__(self._nodeid)
            self._cuindex += 1  # Next Iteration
            return nodeobject
        else:
            raise StopIteration()

    @property
    def _nodeid(self):
        """Node ID."""
        return self._model.getObjectId(ObjectType.NODE.value, self._cuindex)


class Node(object):
    """
    Node Methods.

    :param object model: Open Model Instance
    :param str nodeid: Node ID

    Examples:

    >>> from pyswmm import Simulation, Nodes
    >>>
    >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
    ...     j1 = Nodes(sim)["J1"]
    ...     print(j1.invert_el)
    ...     for step in sim:
    ...         print(j1.depth)
    0.0
    """

    def __init__(self, model, nodeid):
        if not model.fileLoaded:
            raise PYSWMMException("SWMM Model Not Open")
        if nodeid not in model.getObjectIDList(ObjectType.NODE.value):
            raise PYSWMMException("ID Not valid")
        self._model = model
        self._nodeid = nodeid

    # --- Get Parameters
    # -------------------------------------------------------------------------
    @property
    def nodeid(self):
        """
        Get Node ID.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     print(j1.nodeid)
        J1
        """
        return self._nodeid

    def is_junction(self):
        """
        Check if node is a Junction Type.

        :return: is junction
        :rtype: bool

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     print(j1.is_junction())
        True
        """
        return self._model.getNodeType(self._nodeid) is shared_enum.NodeType.JUNCTION

    def is_outfall(self):
        """
        Check if node is a Outfall Type.

        :return: is outfall
        :rtype: bool

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     print(j1.is_outfall())
        True
        """
        return self._model.getNodeType(self._nodeid) is shared_enum.NodeType.OUTFALL

    def is_storage(self):
        """
        Check if node is a Storage Type.

        :return: is storage
        :rtype: bool

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     print(j1.is_storage())
        True
        """
        return self._model.getNodeType(self._nodeid) is shared_enum.NodeType.STORAGE

    def is_divider(self):
        """
        Check if node is a Divider Type.

        :return: is divider
        :rtype: bool

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     print(j1.is_divider())
        True
        """
        return self._model.getNodeType(self._nodeid) is shared_enum.NodeType.DIVIDER

    @property
    def invert_elevation(self):
        """
        Get/set node invert elevation.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     print(j1.invert_elevation)
        0.1

        Setting the value

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     print(j1.invert_elevation)
        ...     j1.invert_elevation = 0.2
        ...     print(j1.invert_elevation)
        0.1
        0.2
        """
        return self._model.getNodeParam(self._nodeid,
                                        NodeParams.invertElev.value)

    @invert_elevation.setter
    def invert_elevation(self, param):
        """Set Node Invert Elevation."""
        self._model.setNodeParam(self._nodeid, NodeParams.invertElev.value,
                                 param)

    @property
    def full_depth(self):
        """
        Get node full depth (Physical Depth of manhole).

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     print(j1.full_depth)
        10

        Setting the value

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     print(j1.full_depth)
        ...     j1.full_depth = 50
        ...     print(j1.full_depth)
        10
        50
        """
        return self._model.getNodeParam(self._nodeid,
                                        NodeParams.fullDepth.value)

    @full_depth.setter
    def full_depth(self, param):
        """Set Node Full Depth."""
        self._model.setNodeParam(self._nodeid, NodeParams.fullDepth.value,
                                 param)

    @property
    def surcharge_depth(self):
        """
        Get/set node surcharge depth.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     print(j1.surcharge_depth)
        10

        Setting the value

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     print(j1.surcharge_depth)
        ...     j1.surcharge_depth = 50
        ...     print(j1.surcharge_depth)
        10
        50
        """
        return self._model.getNodeParam(self._nodeid,
                                        NodeParams.surDepth.value)

    @surcharge_depth.setter
    def surcharge_depth(self, param):
        """Set Node Surcharge Depth."""
        self._model.setNodeParam(self._nodeid, NodeParams.surDepth.value,
                                 param)

    @property
    def ponding_area(self):
        """
        Get/set node ponding area.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     print(j1.ponding_area)
        0

        Setting the value

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     print(j1.ponding_area)
        ...     j1.ponding_area = 50
        ...     print(j1.ponding_area)
        0
        50
        """
        return self._model.getNodeParam(self._nodeid,
                                        NodeParams.pondedArea.value)

    @ponding_area.setter
    def ponding_area(self, param):
        """Set Node Ponding Area."""
        self._model.setNodeParam(self._nodeid, NodeParams.pondedArea.value,
                                 param)

    @property
    def initial_depth(self):
        """
        Get/set node initial depth.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     print(j1.initial_depth)
        0

        Setting the value

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     print(j1.initial_depth)
        ...     j1.initial_depth = 1
        ...     print(j1.initial_depth)
        0
        1
        """
        return self._model.getNodeParam(self._nodeid,
                                        NodeParams.initDepth.value)

    @initial_depth.setter
    def initial_depth(self, param):
        """Set Node Initial Depth."""
        self._model.setNodeParam(self._nodeid, NodeParams.initDepth.value,
                                 param)

    @property
    def total_inflow(self):
        """
        Get Node Results for Total Inflow Rate.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     for step in sim:
        ...         print(j1.total_inflow)
        0
        1.2
        1.5
        1.9
        1.2
        """
        return self._model.getNodeResult(self._nodeid,
                                         NodeResults.totalinflow.value)

    @property
    def total_outflow(self):
        """
        Get Node Results for Total Outflow Rate.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     for step in sim:
        ...         print(j1.total_outflow)
        0
        1.2
        1.5
        1.9
        1.2
        """
        return self._model.getNodeResult(self._nodeid,
                                         NodeResults.outflow.value)

    @property
    def losses(self):
        """
        Get Node Results for Losses Rate (Evap and Exfiltration).

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     for step in sim:
        ...         print(j1.losses)
        0
        0.01
        0.01
        0.01
        0.01
        """
        return self._model.getNodeResult(self._nodeid,
                                         NodeResults.losses.value)

    @property
    def volume(self):
        """
        Get Node Results for Volume.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     for step in sim:
        ...         print(j1.volume)
        0
        1.2
        1.5
        1.9
        1.2
        """
        return self._model.getNodeResult(self._nodeid,
                                         NodeResults.newVolume.value)

    @property
    def flooding(self):
        """
        Get Node Results for Flooding Rate.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     for step in sim:
        ...         print(j1.flooding)
        0
        0
        0.01
        0
        0
        """
        return self._model.getNodeResult(self._nodeid,
                                         NodeResults.overflow.value)

    @property
    def depth(self):
        """
        Get Node Results for Depth.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     for step in sim:
        ...         print(j1.depth)
        0
        0.5
        0.51
        0.52
        0.49
        """
        return self._model.getNodeResult(self._nodeid,
                                         NodeResults.newDepth.value)

    @property
    def head(self):
        """
        Get Node Results for Head.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     for step in sim:
        ...         print(j1.head)
        10
        10.5
        10.51
        10.52
        10.49
        """
        return self._model.getNodeResult(self._nodeid,
                                         NodeResults.newHead.value)

    @property
    def lateral_inflow(self):
        """
        Get Node Results for lateral Inflow rate.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     for step in sim:
        ...         print(j1.lateral_inflow)
        0
        0.25
        0.25
        0.3
        0.4
        """
        return self._model.getNodeResult(self._nodeid,
                                         NodeResults.newLatFlow.value)

    @property
    def hydraulic_retention_time(self):
        """
        Get Node Results for hydraulic retention time.

        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Parameter Value
        :rtype: float
        """
        return self._model.getNodeResult(self._nodeid,
                                         NodeResults.hyd_res_time.value)

    def generated_inflow(self, inflowrate):
        """
        Generate and Set a Node Inflow Rate.

        The value is held constant in the model until it is redefined.
        Generated inflows work like any SWMM inflow.  This does not
        introduce any continuity errors since all flows is counted as
        an inflow.

        :param float inflowrate: Inflow Rate

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     for step in sim:
        ...         j1.generated_inflow(9)
        >>>
        """
        self._model.setNodeInflow(self._nodeid, inflowrate)

    @property
    def cumulative_inflow(self):
        """
        Get Cumulative Node Loading.

        :return: Cumulative Volume
        :rtype: float
        """
        value = self._model.node_inflow(self.nodeid)
        return value

    @property
    def pollut_quality(self):
        """
        Get Current Water Quality Values for a Node.
        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Group of Water Quality Values.
        :rtype: dict

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/test_pollutants.inp') as sim:
        ...     J1 = Nodes(sim)["J1"]
        ...     for step in sim:
        ...         print(J1.pollut_quality)
        {'test-pollutant': 0.0}
        {'test-pollutant': 120.0}
        {'test-pollutant': 120.0}
        """
        out_dict = {}
        pollut_ids = self._model.getObjectIDList(ObjectType.POLLUT.value)
        quality_array = self._model.getNodePollut(self._nodeid,
                                                      NodePollut.nodeQual.value)

        for ind in range(len(pollut_ids)):
            out_dict[pollut_ids[ind]] = quality_array[ind]

        return out_dict

    @pollut_quality.setter
    def pollut_quality(self, args):
        """
        Set Current Node Water Quality Value
        
        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/test_pollutants.inp') as sim:
        ...     Tank = Nodes(sim)["Tank"]
        ...     for step in sim:
        ...         Tank.pollut_quality = ('test-pollutant', 100)
                    print(Tank.pollut_quality)
        {'test-pollutant': 100.0}
        {'test-pollutant': 100.0}
        {'test-pollutant': 100.0}
        """
        pollutant_ID, pollutant_value = args
        
        self._model.setNodePollut(self._nodeid, pollutant_ID, pollutant_value)

    @property
    def inflow_quality(self):
        """
        Get Current Inflow Water Quality Values for a Node.
        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Group of Water Quality Values.
        :rtype: dict

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/test_pollutants.inp') as sim:
        ...     J1 = Nodes(sim)["J1"]
        ...     for step in sim:
        ...         print(J1.inflow_quality)
        {'test-pollutant': 0.0}
        {'test-pollutant': 120.0}
        {'test-pollutant': 120.0}
        """
        out_dict = {}
        pollut_ids = self._model.getObjectIDList(ObjectType.POLLUT.value)
        quality_array = self._model.getNodePollut(self._nodeid,
                                                      NodePollut.inflowQual.value)

        for ind in range(len(pollut_ids)):
            out_dict[pollut_ids[ind]] = quality_array[ind]

        return out_dict

    @property
    def reactor_quality(self):
        """
        Get Current Water Quality Values for the mixed reactor inside a Node.
        If Simulation is not running this method will raise a warning and
        return 0.

        :return: Group of Water Quality Values.
        :rtype: dict

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/test_pollutants.inp') as sim:
        ...     J1 = Nodes(sim)["J1"]
        ...     for step in sim:
        ...         print(J1.reactor_quality)
        {'test-pollutant': 0.0}
        {'test-pollutant': 120.0}
        {'test-pollutant': 120.0}
        """
        out_dict = {}
        pollut_ids = self._model.getObjectIDList(ObjectType.POLLUT.value)
        quality_array = self._model.getNodePollut(self._nodeid,
                                                      NodePollut.reactorQual.value)

        for ind in range(len(pollut_ids)):
            out_dict[pollut_ids[ind]] = quality_array[ind]

        return out_dict

    @property
    def statistics(self):
        """
        Node Statistics. The stats returned are rolling/cumulative.
        Indeces are as follows:

        +-------------------------+
        | average_depth           |
        +-------------------------+
        | max_depth               |
        +-------------------------+
        | max_depth_date          |
        +-------------------------+
        | max_report_depth        |
        +-------------------------+
        | flooding_volume         |
        +-------------------------+
        | flooding_duration       |
        +-------------------------+
        | surcharge_duration      |
        +-------------------------+
        | courant_crit_duration   |
        +-------------------------+
        | lateral_inflow_vol      |
        +-------------------------+
        | peak_lateral_inflowrate |
        +-------------------------+
        | peak_total_inflow       |
        +-------------------------+
        | peak_flooding_rate      |
        +-------------------------+
        | max_ponded_volume       |
        +-------------------------+
        | max_inflow_date         |
        +-------------------------+
        | max_flooding_date       |
        +-------------------------+

        :return: Group of Stats
        :rtype: dict
        """
        return self._model.node_statistics(self.nodeid)


class Outfall(Node):
    """
    Outfall Object: Subclass of Node Object.
    """

    def __init__(self):
        super(Outfall, self).__init__()

    @property
    def outfall_statistics(self):
        """
        Outfall Stats. The stats returned are rolling/cumulative.
        Indeces are as follows:

        +-------------------+
        | average_flowrate  |
        +-------------------+
        | peak_flowrate     |
        +-------------------+
        | pollutant_loading |
        +-------------------+
        | total_periods     |
        +-------------------+

        :return: Group of Stats
        :rtype: list
        """
        return self._model.outfall_statistics(self.nodeid)

    def outfall_stage(self, stage):
        """
        Generate and Set an Outfall Stage (head).

        The value is held constant in the model until it is redefined.
        Using the function overrides the mechanism within SWMM that would
        internerally set the outfall stage.  This does not
        introduce any continuity errors since all flows is counted as
        an inflow.

        :param float stage: Outfall Stage (Head)

        Examples:

        >>> from pyswmm import Simulation, Nodes
        >>>
        >>> with Simulation('tests/data/model_weir_setting.inp') as sim:
        ...     j1 = Nodes(sim)["J1"]
        ...     for step in sim:
        ...         j1.outfall_stage(9)
        """
        self._model.setOutfallStage(self._nodeid, stage)


class Storage(Node):
    """
    Storage Object: Subclass of Node Object.
    """

    def __init__(self):
        super(Storage, self).__init__()

    @property
    def storage_statistics(self):
        """
        Storage Stats. The stats returned are rolling/cumulative.
        Indeces are as follows:

        +----------------+
        | initial_volume |
        +----------------+
        | average_volume |
        +----------------+
        | max_volume     |
        +----------------+
        | peak_flowrate  |
        +----------------+
        | evap_loss      |
        +----------------+
        | exfil_loss     |
        +----------------+
        | max_vol_date   |
        +----------------+

        :return: Group of Stats
        :rtype: list
        """
        return self._model.storage_statistics(self.nodeid)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2021 Jennifer Wu
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
import pytest

from pyswmm import Simulation
from pyswmm import Output, SubcatchSeries, NodeSeries, LinkSeries, SystemSeries
from pyswmm.tests.data import MODEL_WEIR_SETTING_PATH
from pyswmm.errors import OutputException

from swmm.toolkit.shared_enum import LinkAttribute, NodeAttribute, SubcatchAttribute, SystemAttribute
from datetime import datetime


def test_output_unknown_object_id():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        for step in sim:
            pass

    with Output(MODEL_WEIR_SETTING_PATH.replace('inp', 'out')) as out:
        with pytest.raises(OutputException):
            flow_rate = out.link_series('C4', LinkAttribute.FLOW_RATE)


def test_output_invalid_time():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        for step in sim:
            pass

    with Output(MODEL_WEIR_SETTING_PATH.replace('inp', 'out')) as out:
        with pytest.raises(OutputException):
            flow_rate = out.link_series(
                'C3', LinkAttribute.FLOW_RATE, datetime(
                    2015, 10, 1, 15))


def test_output_with():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        for step in sim:
            pass

    with Output(MODEL_WEIR_SETTING_PATH.replace('inp', 'out')) as out:
        assert len(out.subcatchments) == 3
        assert len(out.nodes) == 5
        assert len(out.links) == 4
        assert len(out.pollutants) == 0

        # access with output methods
        flow_rate = out.link_series('C3', LinkAttribute.FLOW_RATE)
        times = list(flow_rate.keys())
        assert times[0] == datetime(2015, 11, 1, 14, 1)
        assert times[-1] == datetime(2015, 11, 4)
        assert len(flow_rate) == 3480

        subset_flow_rate = out.link_series(
            'C3', LinkAttribute.FLOW_RATE, datetime(
                2015, 11, 1, 15))
        subset_times = list(subset_flow_rate.keys())
        assert subset_times[0] == datetime(2015, 11, 1, 15)
        assert subset_times[-1] == datetime(2015, 11, 4)

        subset_flow_rate = out.link_series(
            'C3', LinkAttribute.FLOW_RATE, datetime(
                2015, 11, 2, 15), datetime(
                2015, 11, 3, 15))
        subset_times = list(subset_flow_rate.keys())
        assert subset_times[0] == datetime(2015, 11, 2, 15)
        assert subset_times[-1] == datetime(2015, 11, 3, 14, 59)

        assert len(out.node_series('J1', NodeAttribute.TOTAL_INFLOW)) == 3480
        assert len(
            out.subcatch_series(
                'S1',
                SubcatchAttribute.RUNOFF_RATE)) == 3480
        assert len(out.system_series(SystemAttribute.EVAP_INFIL_LOSS)) == 3480

        assert len(
            out.subcatch_attribute(
                SubcatchAttribute.RUNOFF_RATE,
                0)) == 3
        assert len(out.node_attribute(NodeAttribute.HYDRAULIC_HEAD, 0)) == 5
        assert len(out.link_attribute(LinkAttribute.FLOW_RATE, 0)) == 4
        # waiting for function to be fixed
        # assert len(out.system_attribute('air_temp', 0)) == 1

        # no pollutant
        assert len(out.subcatch_result('S1', 0)) == len(SubcatchAttribute) - 1
        assert len(out.node_result('J1', 0)) == len(NodeAttribute) - 1
        assert len(out.link_result('C1', 0)) == len(LinkAttribute) - 1
        assert len(out.system_result(0)) == len(SystemAttribute)


def test_output():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        for step in sim:
            pass

    out = Output(MODEL_WEIR_SETTING_PATH.replace('inp', 'out'))
    out.open()
    assert len(out.subcatchments) == 3
    assert len(out.nodes) == 5
    assert len(out.links) == 4
    assert len(out.pollutants) == 0
    flow_rate = out.link_series('C3', 'flow_rate')
    times = list(flow_rate.keys())
    assert times[0] == datetime(2015, 11, 1, 14, 1)
    assert times[-1] == datetime(2015, 11, 4)
    assert len(flow_rate) == 3480
    out.close()


def test_timeseries_abstraction():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        for step in sim:
            pass

    with Output(MODEL_WEIR_SETTING_PATH.replace('inp', 'out')) as out:
        for attr in SubcatchAttribute:
            series = getattr(SubcatchSeries(out)['S1'], attr.name.lower())
            assert len(series) == 3480
        for attr in NodeAttribute:
            series = getattr(NodeSeries(out)['J1'], attr.name.lower())
            assert len(series) == 3480
        for attr in LinkAttribute:
            series = getattr(LinkSeries(out)['C1:C2'], attr.name.lower())
            assert len(series) == 3480
        for attr in SystemAttribute:
            series = getattr(SystemSeries(out), attr.name.lower())
            assert len(series) == 3480



from pyswmm import Simulation, RainGages, SystemStats
from pyswmm.tests.data import MODEL_RAIN
from pyswmm.toolkitapi import RainGageResults


def test_api_rainfall():
    with Simulation(MODEL_RAIN) as sim:
        check = 0

        sim._model.setGagePrecip("Gage1", 10.00)
        for ind, step in enumerate(sim):
            sim._model.setGagePrecip("Gage1", 10.00)
            x = sim._model.getGagePrecip("Gage1", RainGageResults.total_precip)

            if round(x, 2) == 10.0:
                check += 1
    assert(check == 718)


def test_rainfall():
    with Simulation(MODEL_RAIN) as sim:
        assert "DUMMY_RG" not in RainGages(sim)
        rg = RainGages(sim)["Gage1"]
        assert(rg.raingageid == "Gage1")

        sim.step_advance(3600)
        for ind, step in enumerate(sim):
            if 0 < ind < 5:
                assert(rg.total_precip == 1)
                assert(rg.rainfall == 1)
                assert(rg.snowfall == 0)

            if ind == 5:
                rg.total_precip = 10

            if ind >= 6:
                assert(int(rg.total_precip) == 10)
                assert(int(rg.rainfall) == 10)
                assert(int(rg.snowfall) == 0)

        stats = SystemStats(sim)
        assert(int(stats.runoff_stats['rainfall']) == 65)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014 Bryant E. McDonnell
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------

import pytest

# Local imports
from pyswmm import Simulation, SystemStats
from pyswmm.swmm5 import PySWMM, SWMMException
from pyswmm.tests.data import MODEL_WEIR_SETTING_PATH


def test_system_flow_routing():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        stats = SystemStats(sim)

        print(r"\n\n\System\n")
        with pytest.raises(Exception):
            print(stats.routing_stats)

        sim.step_advance(1200)
        for ind, step in enumerate(sim):
            print(stats.routing_stats)


def test_system_runoff_routing():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        stats = SystemStats(sim)

        print("\n\nRunoff Routing\n\n")
        with pytest.raises(Exception):
            print(stats.runoff_stats)

        sim.step_advance(1200)
        for ind, step in enumerate(sim):
            print(stats.runoff_stats)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014 Bryant E. McDonnell
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
"""SWMM5 test models."""

# Standard library imports
import os
import sys
DATA_PATH = os.path.abspath(os.path.dirname(__file__))

# Test models paths
MODEL_NODE_INFLOWS_PATH = os.path.join(DATA_PATH, 'model_node_inflows.inp')
MODEL_PUMP_SETTINGS_PATH = os.path.join(DATA_PATH, 'model_pump_setting.inp')
MODEL_TOOLKIT_UNITS_PATH = os.path.join(DATA_PATH, 'model_toolkit_units.inp')
MODEL_WEIR_SETTING_PATH = os.path.join(DATA_PATH, 'model_weir_setting.inp')
MODEL_FULL_FEATURES_PATH = os.path.join(DATA_PATH, 'model_full_features.inp')
MODEL_STORAGE_PUMP = os.path.join(DATA_PATH, 'model_storage_pump.inp')
MODEL_STORAGE_PUMP_MGD = os.path.join(DATA_PATH, 'model_storage_pump_MGD.inp')
MODEL_POLLUTANTS_PATH = os.path.join(DATA_PATH, 'model_pollutants.inp')
MODEL_POLLUTANTS_PATH_2 = os.path.join(DATA_PATH, 'model_pollutants_2.inp')
MODEL_POLLUTANTS_PATH_3 = os.path.join(DATA_PATH, 'model_constantinflow_constanteffluent.inp')
MODEL_POLLUTANTS_SETTERS_PATH = os.path.join(DATA_PATH, 'model_pollutants_setters.inp')
MODEL_RAIN = os.path.join(DATA_PATH, 'model_rain.inp')
MODEL_LIDS_PATH = os.path.join(DATA_PATH, 'model_lids.inp')
MODEL_BAD_INPUT_PATH_1 = os.path.join(DATA_PATH, 'model_bad_input_1.inp')
MODEL_SUBCATCH_STATS_PATH = os.path.join(DATA_PATH, 'model_subcatch_stats.inp')

WIN_SWMM_LIB_PATH = os.path.join(DATA_PATH, '..\\..\\lib\\windows',
                                 'swmm5.dll')
if sys.maxsize > 2**32:
    WIN_SWMM_LIB_PATH = os.path.join(DATA_PATH, '..\\..\\lib\\windows',
                                     'swmm5-x64.dll')
else:
    WIN_SWMM_LIB_PATH = os.path.join(DATA_PATH, '..\\..\\lib\\windows',
                                     'swmm5.dll')



"""
This module implements monkey patches to the swmm.toolkit.solver module.

As the SWMM API is expanded and/or the EPA updates the engine, pyswmm should grow to 
accomodate the new features. However, sometimes the API changes in a way that is not 
backwards compatible. In these cases, we can use monkey patches to add the new features 
to older versions of the toolkit.

Those monkey patches may be python implementations of the new features or they
may be a wrapper around the new toolkit function api that raises an error if the
user tries to use the new feature with an older version of the toolkit.
"""

from typing import Callable
import warnings

import packaging.version
from swmm.toolkit import __version__ as _tk_version
from swmm.toolkit import solver
from pyswmm.swmm5 import PySWMM

# %% ###########################
# region MonkeyPatchTypes ######
################################
class ToolkitVersionException(Exception):
    pass

class MonkeyPatchWarning(Warning):
    pass

class _monkey_patch:
    """
    A class to store and implement monkey patches
    """
    def __init__(
            self,
            function_name: str,
            swmm_toolkit_minimum_version: str,
            object_to_patch: object,
            alternative_function: Callable = None,
            warning_message: str = None,
        ):
        self.function_name = function_name
        self.swmm_toolkit_minimum_version = swmm_toolkit_minimum_version
        self._alternative_function = alternative_function
        self.warning_message = warning_message
        self._object_to_patch = object_to_patch

    def _toolkit_error_func(self,*args,**kwargs):
        """A default monkey patch function that raises a toolkit version error"""        
        raise ToolkitVersionException(
            f'SWMM version must be at least {self.swmm_toolkit_minimum_version}. '
            f"The currently installed version is {_tk_version}"
        )
    
    @property
    def alt_function(self):
        """Return the alternative function or the default toolkit error function 
        depending on availability"""
        if self._alternative_function is None:
            return self._toolkit_error_func
        else:
            return self._alternative_function
    
    @property
    def force_patch(self):
        """
        always patch if the alternative function is not None

        There might be a case where we don't necesarily want to patch the function but only
        want to issue a warning at runtime (e.g. if api doen't change but as expanded capabilities in a newer version)
        """
        return True if self._alternative_function is not None else False
    
    def patch(self):
        """Apply monkey patch to module if toolkit version is less than minimum version"""
        if packaging.version.parse(_tk_version) < packaging.version.parse(self.swmm_toolkit_minimum_version):
            if self.warning_message is not None:
                warnings.warn(self.warning_message,MonkeyPatchWarning)
            
            if not hasattr(self._object_to_patch,self.function_name) or self.force_patch:
                setattr(self._object_to_patch,self.function_name,self.alt_function) 

# endregion MonkeyPatchTypes ####

# %% ##############################
# region PySWMMObjectPatches ######
###################################

def swmm_stride(self, advanceSeconds):
    """
    This function allows for user defined stride length to advance
    the model simulation by a defined time.  This is useful when control
    rules are managed externally by PySWMM. Instead of evaluating rules
    every routing step, instead the simulation can be advanced further
    in time before the PySWMM can intervene. When a 0 is returned, the
    simulation period has reached the end.

    :param int advanceSeconds: Number seconds to advance the simulation
                                forward.
    :return: Current simulation time after a stride in decimal days (float)
    :rtype: float

    Examples:

    >>> swmm_model = PySWMM(r'\\.inp',r'\\.rpt',r'\\.out')
    >>> swmm_model.swmm_open()
    >>> swmm_model.swmm_start(True)
    >>> while(True):
    ...     time = swmm_model.swmm_stride(600)
    ...     if (time <= 0.0): break
    >>>
    >>> swmm_model.swmm_end()
    >>> swmm_model.swmm_report()
    >>> swmm_model.swmm_close()
    """
    ctime = self.curSimTime
    secPday = 3600.0 * 24.0
    advanceDays = advanceSeconds / secPday
    eps = advanceDays * 0.00001
    elapsed_time = 0

    while self.curSimTime <= ctime + advanceDays - eps:
        elapsed_time = solver.swmm_step()
        if elapsed_time == 0:
            return 0.0
        self.curSimTime = elapsed_time

    return elapsed_time

# endregion PySWMMObjectPatches ####


# Patches
_patches = [
    _monkey_patch(function_name = "swmm_hotstart",swmm_toolkit_minimum_version="0.15.0",object_to_patch=solver),
    _monkey_patch(function_name = "swmm_stride" ,swmm_toolkit_minimum_version="0.15.0",object_to_patch=PySWMM,alternative_function=swmm_stride),

]
def patch():    
    # patch solver
    for monkey_patch in _patches:        
        monkey_patch.patch()


# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014 Bryant E. McDonnell
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------

# Standard library imports
from random import randint

# Local imports
import pyswmm.toolkitapi as tka
from pyswmm import Links, Nodes, Simulation, SimulationPreConfig
from pyswmm.errors import MultiSimulationError
from pyswmm.tests.data import MODEL_WEIR_SETTING_PATH
import pytest
import os


def test_simulation_1():
    sim = Simulation(MODEL_WEIR_SETTING_PATH)
    print(f'system units: {sim.system_units}')
    print(f'swmm version: {sim._model.swmm_getVersion()}')
    allow_ponding = sim._model.getSimOptionSetting(
        tka.SimAnalysisSettings.AllowPonding)
    routing_step = sim._model.getSimAnalysisSetting(
        tka.SimulationParameters.RouteStep)
    print(f'analysis setting: {allow_ponding}')
    print(f'analysis param: {routing_step}')
    assert sim.system_units == 'US'
    assert not allow_ponding
    assert routing_step == 1
    for ind, step in enumerate(sim):
        print(sim.current_time)
        sim.step_advance(randint(300, 900))

    sim.close()


def test_simulation_2():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        for ind, step in enumerate(sim):
            print(sim.current_time)
            sim.step_advance(randint(300, 900))


def test_simulation_3():
    sim = Simulation(MODEL_WEIR_SETTING_PATH)
    sim.execute()


def test_simulation_4():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        for ind, step in enumerate(sim):
            print(sim.percent_complete)
            sim.step_advance(randint(300, 900))


def test_simulation_iter():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        c1c2 = Links(sim)["C1:C2"]

        sim.step_advance(300)
        for ind, step in enumerate(sim):
            print(c1c2.flow, c1c2.target_setting)
            if c1c2.flow > 9.19:
                c1c2.target_setting = 0.9


def test_simulation_9():
    '''Modified test to use "before_start" '''
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        J1 = Nodes(sim)["J1"]

        def init_function():
            J1.initial_depth = 15

        sim.add_before_start(init_function)
        for ind, step in enumerate(sim):
            if ind == 0:
                assert J1.depth > 14


def test_simulation_callback_1():
    LIST = []
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        LIST.append("OPENED")

        def before_start1():
            LIST.append("before_start1")

        sim.add_before_start(before_start1)

        def after_start1():
            LIST.append("after_start1")

        sim.add_after_start(after_start1)

        def before_step1():
            if "before_step1" not in LIST:
                LIST.append("before_step1")

        sim.add_before_step(before_step1)

        def after_step1():
            if "after_step1" not in LIST:
                LIST.append("after_step1")

        sim.add_after_step(after_step1)

        def before_end1():
            if "before_end1" not in LIST:
                LIST.append("before_end1")

        sim.add_before_end(before_end1)

        def after_end1():
            if "after_end1" not in LIST:
                LIST.append("after_end1")

        sim.add_after_end(after_end1)

        def after_close1():
            if "after_close1" not in LIST:
                LIST.append("after_close1")

        sim.add_after_close(after_close1)

        for ind, step in enumerate(sim):
            if ind == 0:
                LIST.append("SIM_STEP")

    assert LIST == [
        "OPENED", "before_start1", "after_start1", "before_step1",
        "after_step1", "SIM_STEP", "before_end1", "after_end1", "after_close1"
    ]
    print(LIST)


def test_simulation_terminate():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        i = 0
        for ind, step in enumerate(sim):
            i += 1
            if ind == 10:
                sim.terminate_simulation()
        assert (i == 11)


def test_hotstart():
    HSF_PATH = MODEL_WEIR_SETTING_PATH.replace('.inp', '.hsf')
    if os.path.exists(HSF_PATH):
        os.remove(HSF_PATH)

    # test saving hotstart works
    assert not os.path.exists(HSF_PATH)

    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        J1 = Nodes(sim)["J1"]
        for ind, step in enumerate(sim):
            if ind == 10:
                sim.save_hotstart(HSF_PATH)
                J1_dep = J1.depth
                break

    assert os.path.exists(HSF_PATH)

    # test loading hotstart works and that data matches
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        def store_J1_depth_before_step():
            sim.J1_depth = Nodes(sim)["J1"].depth
        sim.add_before_step(store_J1_depth_before_step)
        sim.use_hotstart(HSF_PATH)

        for ind, step in enumerate(sim):
            break
    assert sim.J1_depth == pytest.approx(J1_dep, 0.00001)

def test_pre_simulation_config():
    sim_preconfig = SimulationPreConfig()
    sim_preconfig.filename_suffix = "_a"

    sim_preconfig.add_update_by_token("SUBCATCHMENTS", "S1", 2, "J2")
    sim_preconfig.add_update_by_token(
        "TIMESERIES", "SCS_24h_Type_I_1in", 2, 2.0, 5)

    with Simulation(MODEL_WEIR_SETTING_PATH,
                    sim_preconfig=sim_preconfig) as sim:
        pass

    with open(MODEL_WEIR_SETTING_PATH.replace(".inp", "_a.inp"), 'r') as fl:
        for ind, ln in enumerate(fl):
            if ind == 55:
                compare = ['S1', 'SCS_24h_Type_I_1in',
                           'J2', '1', '100', '500', '0.5', '0']
                ln = ln.strip()
                ln = ln.split()
                assert (ln == compare)
            if ind == 137:
                compare = ['SCS_24h_Type_I_1in', '1:15', '2.0']
                ln = ln.strip()
                ln = ln.split()
                assert (ln == compare)

def test_multi_sim_exception():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        with pytest.raises(MultiSimulationError):
            with Simulation(MODEL_WEIR_SETTING_PATH) as sim2:
                pass
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim3:
        pass

def test_states():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        assert (sim.sim_is_open == True)
    assert(sim.sim_is_open == False)

    with Simulation(MODEL_WEIR_SETTING_PATH) as sim3:
        assert (sim.sim_is_started == False)
        for step in sim:
            assert (sim.sim_is_started == True)

    sim4 = Simulation(MODEL_WEIR_SETTING_PATH)
    assert (sim4.sim_is_open == True)
    sim4.close()
    assert (sim4.sim_is_open == False)

    sim5 = Simulation(MODEL_WEIR_SETTING_PATH)
    assert (sim5.sim_is_open == True)
    sim5.execute()
    assert (sim5.sim_is_open == False)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014 Bryant E. McDonnell
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------

# Standard library imports
import sys

# Local imports
from pyswmm import Simulation
from pyswmm.tests.data import MODEL_WEIR_SETTING_PATH
from pyswmm.utils.fixtures import get_model_files
from pyswmm.swmm5 import PySWMM

# def test_engine_version():
# if sys.platform == 'darwin':
# pyswmm.lib.use("swmm5.dylib")
# elif sys.platform.startswith('linux'):
# pyswmm.lib.use("swmm5.so")
# else:
# if sys.maxsize > 2**32:
# pyswmm.lib.use("swmm5-x64.dll")
# else:
# pyswmm.lib.use("swmm5.dll")
##
##    sim = Simulation(MODEL_WEIR_SETTING_PATH)
# print(sim.engine_version)

def test_swmm_stride():
    swmmobject = PySWMM(*get_model_files(MODEL_WEIR_SETTING_PATH))
    swmmobject.swmm_open()
    swmmobject.swmm_start()

    i = 0
    while (True):
        time = swmmobject.swmm_stride(600)
        i += 1

        if i == 10:
            break
    print(time)
    assert time == (600 * 10)/86400
    swmmobject.swmm_end()
    swmmobject.swmm_close()

def test_runoff_error():
    sim = Simulation(MODEL_WEIR_SETTING_PATH)
    sim.execute()
    print(sim.runoff_error)
    assert sim.runoff_error >= 0.0


def test_flow_routing_error():
    sim = Simulation(MODEL_WEIR_SETTING_PATH)
    sim.execute()
    print(sim.flow_routing_error)
    assert sim.flow_routing_error >= 0.0


def test_quality_error():
    sim = Simulation(MODEL_WEIR_SETTING_PATH)
    sim.execute()
    print(sim.quality_error)
    assert sim.quality_error >= 0.0




# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014 Bryant E. McDonnell
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------

# Local imports
from pyswmm import Node, Nodes, Simulation
from pyswmm.swmm5 import PySWMM
from pyswmm.tests.data import (MODEL_FULL_FEATURES_PATH,
                               MODEL_NODE_INFLOWS_PATH, MODEL_STORAGE_PUMP,
                               MODEL_STORAGE_PUMP_MGD, MODEL_WEIR_SETTING_PATH)
from pyswmm.utils.fixtures import get_model_files
import pyswmm.toolkitapi as tka
from pytest import approx
UT_PRECISION = 1  # %


def test_nodes_1():
    ''' pytest pyswmm/tests/test_nodes.py -k `test_nodes_1` '''
    swmmobject = PySWMM(MODEL_WEIR_SETTING_PATH)
    swmmobject.swmm_open()
    node = Node(swmmobject, "J1")
    assert(node.invert_elevation == 20.728)


def test_nodes_2():
    ''' pytest pyswmm/tests/test_nodes.py -k `test_nodes_2` '''
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        print("\n\n\nNODES\n")
        assert "DUMMY_LINK" not in Nodes(sim)
        for node in Nodes(sim):
            assert ('J' in node.nodeid)
            node.invert_elevation = 10
            assert (node.invert_elevation == 10)


def test_nodes_3():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        print("\n\n\nNODES\n")
        j1 = Nodes(sim)["J1"]
        assert (j1.is_divider() == False)
        assert (j1.is_junction() == True)
        assert (j1.is_outfall() == False)
        assert (j1.is_storage() == False)


def test_nodes_4():
    swmmobject = PySWMM(*get_model_files(MODEL_WEIR_SETTING_PATH))
    swmmobject.swmm_open()

    print(swmmobject.getNodeParam('J2', tka.NodeParams.invertElev))
    swmmobject.setNodeParam('J2', tka.NodeParams.invertElev, 19)
    print(swmmobject.getNodeParam('J2', tka.NodeParams.invertElev))

    print(swmmobject.getLinkParam('C1:C2', tka.LinkParams.offset1))
    swmmobject.setLinkParam('C1:C2', tka.LinkParams.offset1, 19)
    print(swmmobject.getLinkParam('C1:C2', tka.LinkParams.offset1))

    print(swmmobject.getSubcatchParam('S2', tka.SubcParams.area))
    swmmobject.setSubcatchParam('S2', tka.SubcParams.area, 19)
    print(swmmobject.getSubcatchParam('S2', tka.SubcParams.area))

    swmmobject.swmm_start()
    i = 0
    while (True):
        time = swmmobject.swmm_stride(600)
        i += 1

        if (time <= 0.0):
            break

        if i % 144 == 0:
            print(i)

        swmmobject.setLinkParam('C1:C2', tka.LinkParams.qLimit, 0.25)

    swmmobject.swmm_end()
    swmmobject.swmm_close()
    print("swmm_step() Check Passed")


def test_nodes_5():
    sim = Simulation(MODEL_WEIR_SETTING_PATH)
    print("\n\n\nNODES\n")
    J5 = Nodes(sim)["J5"]

    for ind, step in enumerate(sim):
        if ind == 7:
            J5.generated_inflow(544.0)
        if ind > 8:
            assert (J5.lateral_inflow == approx(543.9, rel=UT_PRECISION))
            if ind % 1000 == 0:
                print(J5.lateral_inflow)
    sim.close()


def test_nodes_6():
    ''' pytest pyswmm/tests/test_nodes.py -k `test_nodes_6` '''
    sim = Simulation(MODEL_WEIR_SETTING_PATH)
    print("\n\n\nNODES\n")
    J5 = Nodes(sim)["J5"]

    for ind, step in enumerate(sim):
        pass

    assert(J5.statistics['peak_total_inflow']
           == approx(0.478, rel=UT_PRECISION))
    assert(J5.statistics['average_depth'] == approx(0.00017, rel=UT_PRECISION))
    assert(J5.statistics['surcharge_duration']
           == approx(0.0, rel=UT_PRECISION))
    assert(J5.statistics['max_ponded_volume'] == approx(0.0, rel=UT_PRECISION))
    assert(J5.statistics['courant_crit_duration']
           == approx(0.0, rel=UT_PRECISION))
    assert(J5.statistics['peak_lateral_inflowrate']
           == approx(0.0, rel=UT_PRECISION))
    assert(J5.statistics['flooding_duration'] == approx(0.0, rel=UT_PRECISION))
    assert(J5.statistics['peak_flooding_rate']
           == approx(0.0, rel=UT_PRECISION))
    assert(J5.statistics['lateral_infow_vol'] == approx(0.0, rel=UT_PRECISION))
    assert(J5.statistics['max_flooding_date']
           == approx(42309, rel=UT_PRECISION))
    assert(J5.statistics['max_depth_date'] == approx(42309, rel=UT_PRECISION))
    assert(J5.statistics['max_inflow_date'] == approx(42309, rel=UT_PRECISION))
    assert(J5.statistics['max_depth'] == approx(0.0292, rel=UT_PRECISION))
    assert(J5.statistics['flooding_volume'] == approx(0.0, rel=UT_PRECISION))
    assert(J5.statistics['max_report_depth']
           == approx(0.0286, rel=UT_PRECISION))

    sim.close()


def test_storage_7():
    ''' pytest pyswmm/tests/test_nodes.py -k `test_storage_7` '''
    sim = Simulation(MODEL_STORAGE_PUMP)
    print("\n\n\nSTORAGE\n")
    STOR1 = Nodes(sim)["SU1"]

    for ind, step in enumerate(sim):
        pass

    stats = STOR1.storage_statistics
    assert (stats['max_volume'] == approx(5000.0, rel=UT_PRECISION))
    assert (stats['peak_flowrate'] == approx(0.0, rel=UT_PRECISION))
    assert (stats['exfil_loss'] == approx(0.0, rel=UT_PRECISION))
    assert (stats['evap_loss'] == approx(0.0, rel=UT_PRECISION))
    assert (stats['average_volume'] == approx(4979.0, rel=UT_PRECISION))
    assert (stats['average_volume'] == approx(4980.0, rel=UT_PRECISION))
    assert (stats['max_vol_date'] == approx(42309.0, rel=UT_PRECISION))
    assert (stats['max_vol_date'] == approx(42310.0, rel=UT_PRECISION))
    assert (stats['initial_volume'] == approx(0.0, rel=UT_PRECISION))

    stats = STOR1.statistics
    assert (stats['peak_total_inflow'] == approx(2.9, rel=UT_PRECISION))
    assert (stats['peak_total_inflow'] == approx(3.1, rel=UT_PRECISION))

    assert (stats['average_depth'] == approx(4.9, rel=UT_PRECISION))
    assert (stats['average_depth'] == approx(5, rel=UT_PRECISION))

    assert (stats['flooding_duration'] == approx(57, rel=UT_PRECISION))
    assert (stats['flooding_duration'] == approx(58, rel=UT_PRECISION))

    assert (stats['peak_flooding_rate'] == approx(2.9, rel=UT_PRECISION))
    assert (stats['peak_flooding_rate'] == approx(3.1, rel=UT_PRECISION))

    assert (stats['lateral_infow_vol'] == approx(0.0, rel=UT_PRECISION))

    assert (stats['max_flooding_date'] == approx(42309, rel=UT_PRECISION))
    assert (stats['max_flooding_date'] == approx(42310, rel=UT_PRECISION))

    assert (stats['max_depth_date'] == approx(42309, rel=UT_PRECISION))
    assert (stats['max_depth_date'] == approx(42310, rel=UT_PRECISION))

    assert (stats['max_inflow_date'] == approx(42309, rel=UT_PRECISION))
    assert (stats['max_inflow_date'] == approx(42310, rel=UT_PRECISION))

    assert (stats['max_depth'] == approx(4.99, rel=UT_PRECISION))
    assert (stats['max_depth'] == approx(5.01, rel=UT_PRECISION))

    assert (stats['flooding_volume'] == approx(621390, rel=UT_PRECISION))
    assert (stats['flooding_volume'] == approx(621393, rel=UT_PRECISION))

    assert (stats['max_report_depth'] == approx(4.99, rel=UT_PRECISION))
    assert (stats['max_report_depth'] == approx(5.01, rel=UT_PRECISION))

    print(STOR1.statistics)

    sim.close()


def test_outfalls_8():
    ''' pytest pyswmm/tests/test_nodes.py -k `test_outfalls_8` '''
    sim = Simulation(MODEL_STORAGE_PUMP)
    print("\n\n\nOUTFALL\n")
    outfall = Nodes(sim)["J3"]

    for ind, step in enumerate(sim):
        pass
    stats = outfall.outfall_statistics
    outfall_cuinflow = outfall.cumulative_inflow
    sim.close()
    assert(stats['total_periods'] == approx(208796, rel=UT_PRECISION))
    assert(stats['pollutant_loading']['test']
           == approx(1756, rel=UT_PRECISION))
    assert(stats['pollutant_loading']['test']
           == approx(1756.2, rel=UT_PRECISION))
    assert(stats['average_flowrate'] == approx(8.9, rel=UT_PRECISION))
    assert(stats['average_flowrate'] == approx(9.0, rel=UT_PRECISION))
    assert(stats['peak_flowrate'] == approx(9.0, rel=UT_PRECISION))
    assert(stats['peak_flowrate'] == approx(9.1, rel=UT_PRECISION))
    assert(outfall_cuinflow == approx(1876800, rel=UT_PRECISION))
    assert(outfall_cuinflow == approx(1876900, rel=UT_PRECISION))


def test_outfalls_8_mgd():
    ''' pytest pyswmm/tests/test_nodes.py -k `test_outfalls_8_mgd` '''
    sim = Simulation(MODEL_STORAGE_PUMP_MGD)
    print("\n\n\nOUTFALL\n")
    outfall = Nodes(sim)["J3"]
    storage = Nodes(sim)["SU1"]
    junction = Nodes(sim)["J2"]

    for ind, step in enumerate(sim):
        pass
    stats = outfall.outfall_statistics
    outfall_cuinflow = outfall.cumulative_inflow
    sim.close()

    assert (stats['total_periods'] == approx(208796, rel=UT_PRECISION))
    assert (stats['pollutant_loading']['test']
            == approx(1305.25, rel=UT_PRECISION))
    assert (stats['pollutant_loading']['test']
            == approx(1305.75, rel=UT_PRECISION))
    assert (stats['average_flowrate'] == approx(4.3, rel=UT_PRECISION))
    assert (stats['average_flowrate'] == approx(4.32, rel=UT_PRECISION))
    assert (stats['peak_flowrate'] == approx(4.33, rel=UT_PRECISION))
    assert (stats['peak_flowrate'] == approx(4.34, rel=UT_PRECISION))
    assert (outfall_cuinflow == approx(1395293, rel=UT_PRECISION))
    assert (outfall_cuinflow == approx(1395299, rel=UT_PRECISION) and
            outfall_cuinflow == approx(1395350, rel=UT_PRECISION))


def test_nodes_10():
    with Simulation(MODEL_NODE_INFLOWS_PATH) as sim:
        J1 = Nodes(sim)["J1"]
        outfall = Nodes(sim)["J3"]

        J1.generated_inflow(4)
        # Below Invert test
        outfall.outfall_stage(0)
        for ind, step in enumerate(sim):
            if ind == 1000:
                assert outfall.head >= outfall.invert_elevation
            if ind == 5000:
                outfall.outfall_stage(7)
            if ind == 5001:
                assert outfall.head == approx(7.00001, rel=UT_PRECISION)
                assert outfall.head == approx(6.99999, rel=UT_PRECISION)

            if ind == 50000:
                outfall.outfall_stage(13.5)

            if ind == 50001:
                assert outfall.head == approx(13.50001, rel=UT_PRECISION)
                assert outfall.head == approx(13.49999, rel=UT_PRECISION)


def test_nodes_11():
    ''' pytest pyswmm/tests/test_nodes.py -k `test_outfalls_8_mgd` '''
    sim = Simulation(MODEL_STORAGE_PUMP_MGD)
    print("\n\n\nOUTFALL\n")
    outfall = Nodes(sim)["J3"]

    for ind, step in enumerate(sim):
        pass
    stats = outfall.outfall_statistics
    outfall_cuinflow = outfall.cumulative_inflow
    sim.close()
    assert(outfall_cuinflow == approx(1395293, rel=UT_PRECISION))
    assert(outfall_cuinflow == approx(1395299, rel=UT_PRECISION) and
           outfall_cuinflow == approx(1395350, rel=UT_PRECISION))



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014 Bryant E. McDonnell
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
from datetime import datetime

# Third party imports
import pytest

# Local imports
from pyswmm import Simulation, Output
from pyswmm.swmm5 import PySWMM, PYSWMMException
from pyswmm.errors import OutputException
from pyswmm.tests.data import MODEL_WEIR_SETTING_PATH, MODEL_BAD_INPUT_PATH_1
from pyswmm.utils.fixtures import get_model_files


def test_error_rpt_out():
    swmmobject = PySWMM(MODEL_WEIR_SETTING_PATH)
    swmmobject.swmm_open()


#    swmmobject.swmmExec()
#    swmmobject.swmm_close()


def test_pyswmm_exception():
    swmmobject = PySWMM(*get_model_files(MODEL_WEIR_SETTING_PATH))
    with pytest.raises(PYSWMMException):
        swmmobject.swmm_open()
        print(swmmobject.fileLoaded)
        swmmobject.swmm_open()
        swmmobject.swmm_close()


def test_swmm_input_error_1():
    with pytest.raises(Exception) as e:
        with Simulation(MODEL_BAD_INPUT_PATH_1) as sim:
            for step in sim:
                pass
    assert(str(e.value).strip() == 'ERROR 200: one or more errors in input file.')


def test_error_output():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        for step in sim:
            pass

    with Output(MODEL_WEIR_SETTING_PATH.replace('inp', 'out')) as out:
        with pytest.raises(OutputException):
            out.node_series('potato', 'total_inflow')

        with pytest.raises(OutputException):
            out.link_series('C3', 'flow_rate', datetime(2021, 1, 1))

        with pytest.raises(OutputException):
            out.link_series('C3', 'flow_rate', datetime(2015, 11, 4), datetime(2021, 1, 1))


# Local imports
from swmm.toolkit.solver import lid_usage_get_flux_rate
from pyswmm import Simulation
from pyswmm import LidControls, LidGroups
from pyswmm.tests.data import MODEL_LIDS_PATH
from pytest import approx
UT_PRECISION = 1  # %


def test_list_lid_controls():
    with Simulation(MODEL_LIDS_PATH) as sim:
        for i, control in enumerate(LidControls(sim)):
            if i == 0:
                assert(str(control) == 'LID')
            if i == 1:
                assert(str(control) == 'Green_LID')


def test_list_lid_groups():
    with Simulation(MODEL_LIDS_PATH) as sim:
        assert "DUMMY_GROUP" not in LidGroups(sim)
        for i, group in enumerate(LidGroups(sim)):
            if i == 0:
                assert('subcatchment {} has {} lid units'.format(group,
                                                                 len(group)) == 'subcatchment 1 has 0 lid units')
            if i == 1:
                assert('subcatchment {} has {} lid units'.format(group,
                                                                 len(group)) == 'subcatchment 2 has 3 lid units')
            if i == 2:
                assert('subcatchment {} has {} lid units'.format(group,
                                                                 len(group)) == 'subcatchment 3 has 0 lid units')
            if i == 3:
                assert('subcatchment {} has {} lid units'.format(group,
                                                                 len(group)) == 'subcatchment 4 has 0 lid units')
            if i == 7:
                assert('subcatchment {} has {} lid units'.format(group,
                                                                 len(group)) == 'subcatchment 8 has 0 lid units')


def test_list_lid_units():
    with Simulation(MODEL_LIDS_PATH) as sim:
        sub_2_lid_units = LidGroups(sim)["2"]
        for i, lid_unit in enumerate(sub_2_lid_units):
            if i == 0:
                assert(lid_unit.subcatchment == '2')
                assert(lid_unit.lid_control == 'LID')
            if i == 1:
                assert(lid_unit.subcatchment == '2')
                assert(lid_unit.lid_control == 'LID')
            if i == 2:
                assert(lid_unit.subcatchment == '2')
                assert(lid_unit.lid_control == 'Green_LID')


def test_lid_group_params():
    with Simulation(MODEL_LIDS_PATH) as sim:
        sub_2_lid_units = LidGroups(sim)["2"]
        sim.step_advance(900)
        for i, step in enumerate(sim):
            if i == 115:
                assert(sim.current_time.strftime(
                    '%Y-%m-%d %H:%M:%S') == '1998-01-02 05:00:00')
                assert(
                    sub_2_lid_units.pervious_area == approx(
                        50000, rel=UT_PRECISION))
                assert(
                    sub_2_lid_units.flow_to_pervious == approx(
                        0, rel=UT_PRECISION))
                assert(
                    sub_2_lid_units.old_drain_flow == approx(
                        0.0008, rel=UT_PRECISION))
                assert(
                    sub_2_lid_units.new_drain_flow == approx(
                        0.0008, rel=UT_PRECISION))


def test_lid_unit_params():
    with Simulation(MODEL_LIDS_PATH) as sim:
        sub_2_lid_units = LidGroups(sim)["2"]
        first_unit = sub_2_lid_units[0]
        assert(first_unit.unit_area == approx(10000, rel=UT_PRECISION))
        assert(first_unit.full_width == approx(20, rel=UT_PRECISION))
        assert(first_unit.initial_saturation == approx(0, rel=UT_PRECISION))
        assert(first_unit.from_impervious == approx(50, rel=UT_PRECISION))
        assert(first_unit.index == 0)
        assert(first_unit.number == 4)
        assert(first_unit.drain_subcatchment == -1)
        assert(first_unit.drain_node == 1)

        for i, step in enumerate(sim):
            if i == 2145:
                assert(
                    first_unit.water_balance.inflow == approx(
                        8.9, rel=UT_PRECISION))
                assert(
                    first_unit.water_balance.evaporation == approx(
                        0, rel=UT_PRECISION))
                assert(
                    first_unit.water_balance.infiltration == approx(
                        0.014, rel=UT_PRECISION))
                assert(
                    first_unit.water_balance.surface_flow == approx(
                        0, rel=UT_PRECISION))
                assert(
                    first_unit.water_balance.drain_flow == approx(
                        3.35, rel=UT_PRECISION))
                assert(
                    first_unit.water_balance.initial_volume == approx(
                        3, rel=UT_PRECISION))
                assert(
                    first_unit.water_balance.final_volume == approx(
                        8.5, rel=UT_PRECISION))
                assert(first_unit.surface.depth == approx(0, rel=UT_PRECISION))
                assert(
                    first_unit.pavement.depth == approx(
                        0, rel=UT_PRECISION))
                assert(
                    first_unit.soil.moisture == approx(
                        0.2, rel=UT_PRECISION))
                assert(first_unit.storage.depth == approx(6, rel=UT_PRECISION))
                assert(first_unit.dry_time == approx(21600, rel=UT_PRECISION))
                assert(
                    first_unit.old_drain_flow == approx(
                        0.0005, rel=UT_PRECISION))
                assert(
                    first_unit.new_drain_flow == approx(
                        0.0005, rel=UT_PRECISION))
                assert(
                    first_unit.surface.flux_rate == approx(
                        0, rel=UT_PRECISION))


def test_lid_control_params():
    with Simulation(MODEL_LIDS_PATH) as sim:
        LID = LidControls(sim)["LID"]
        Green_Lid = LidControls(sim)["Green_LID"]

        assert(LID.surface.thickness == approx(10, rel=UT_PRECISION))
        assert(LID.surface.void_fraction == approx(0.5, rel=UT_PRECISION))
        assert(LID.surface.roughness == approx(0.013, rel=UT_PRECISION))
        assert(LID.surface.slope == approx(1, rel=UT_PRECISION))
        assert(LID.surface.side_slope == approx(5, rel=UT_PRECISION))
        assert(LID.surface.alpha == approx(11.4615, rel=UT_PRECISION))
        assert(LID.can_overflow == False)
        assert(LID.soil.thickness == approx(30, rel=UT_PRECISION))
        assert(LID.soil.porosity == approx(0.5, rel=UT_PRECISION))
        assert(LID.soil.field_capacity == approx(0.2, rel=UT_PRECISION))
        assert(LID.soil.wilting_point == approx(0.1, rel=UT_PRECISION))
        assert(LID.soil.suction_head == approx(3.5, rel=UT_PRECISION))
        assert(LID.soil.k_saturated == approx(5.0, rel=UT_PRECISION))
        assert(LID.soil.k_slope == approx(10, rel=UT_PRECISION))
        assert(LID.storage.thickness == approx(40, rel=UT_PRECISION))
        assert(LID.storage.void_fraction == approx(0.75, rel=UT_PRECISION))
        assert(LID.storage.k_saturated == approx(0.5, rel=UT_PRECISION))
        assert(LID.storage.clog_factor == approx(0.2, rel=UT_PRECISION))
        assert(LID.pavement.thickness == approx(20, rel=UT_PRECISION))
        assert(LID.pavement.void_fraction == approx(0.15, rel=UT_PRECISION))
        assert(LID.pavement.impervious_fraction ==
               approx(0.5, rel=UT_PRECISION))
        assert(LID.pavement.k_saturated == approx(100, rel=UT_PRECISION))
        assert(LID.pavement.clog_factor == approx(8, rel=UT_PRECISION))
        assert(LID.drain.coefficient == approx(0.5, rel=UT_PRECISION))
        assert(LID.drain.exponent == approx(0.5, rel=UT_PRECISION))
        assert(LID.drain.offset == approx(6, rel=UT_PRECISION))
        assert(LID.drain.delay == approx(6, rel=UT_PRECISION))
        # no drainmat
        assert(LID.drain_mat.thickness == approx(0, rel=UT_PRECISION))
        assert(LID.drain_mat.void_fraction == approx(0, rel=UT_PRECISION))
        assert(LID.drain_mat.roughness == approx(0, rel=UT_PRECISION))
        assert(LID.drain_mat.alpha == approx(0, rel=UT_PRECISION))


def test_lid_detailed_report():
    with Simulation(MODEL_LIDS_PATH) as sim:

        subLIDs = LidGroups(sim)

        sub_2_lids = subLIDs["2"]

        # first one defined in .inp
        first_LID_unit_on_sub_2 = sub_2_lids[0]
        # second one defined in .inp
        second_LID_unit_on_sub_2 = sub_2_lids[1]

        assert(first_LID_unit_on_sub_2.number == 4)
        assert(second_LID_unit_on_sub_2.number == 1)

        sim.step_advance(900)
        for i, step in enumerate(sim):
            if i == 20:
                assert(sim.current_time.strftime(
                    '%Y-%m-%d %H:%M:%S') == '1998-01-01 05:15:00')
                assert(
                    first_LID_unit_on_sub_2.surface.inflow == approx(
                        0.47, rel=UT_PRECISION))
                assert(
                    first_LID_unit_on_sub_2.evaporation == approx(
                        0, rel=UT_PRECISION))
                assert(
                    first_LID_unit_on_sub_2.surface.infiltration == approx(
                        0.47, rel=UT_PRECISION))
                assert(
                    first_LID_unit_on_sub_2.pavement.percolation == approx(
                        0.47, rel=UT_PRECISION))
                assert(
                    first_LID_unit_on_sub_2.soil.percolation == approx(
                        0.615, rel=UT_PRECISION))
                assert(
                    first_LID_unit_on_sub_2.storage.exfiltration == approx(
                        0, rel=UT_PRECISION))
                assert(
                    first_LID_unit_on_sub_2.surface.outflow == approx(
                        0, rel=UT_PRECISION))
                assert(
                    first_LID_unit_on_sub_2.storage.drain == approx(
                        0, rel=UT_PRECISION))



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014 Bryant E. McDonnell
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------

# Local imports
from pyswmm import Simulation, Subcatchments
# from pyswmm.swmm5 import PySWMM
from pyswmm.tests.data import MODEL_FULL_FEATURES_PATH, MODEL_WEIR_SETTING_PATH, MODEL_SUBCATCH_STATS_PATH
from pytest import approx

UT_PRECISION = 1  # %


def test_subcatchments_1():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        print("\n\n\nSubcatchments\n")
        S3 = Subcatchments(sim)["S3"]
        print(S3.area)
        print(S3.width)
        print(S3.connection)
        print(S3.percent_impervious)
        print(S3.slope)

        sim.step_advance(300)
        for ind, step in enumerate(sim):
            print(S3.runoff)


def test_subcatchments_2():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:
        assert "DUMMY_SUBC" not in Subcatchments(sim)
        for subcatchment in Subcatchments(sim):
            print(subcatchment)
            print(subcatchment.subcatchmentid)
            print(subcatchment.curb_length)
            subcatchment.curb_length = 10
            print(subcatchment.curb_length)


def test_subcatchments_3():
    with Simulation(MODEL_SUBCATCH_STATS_PATH) as sim:
        S1 = Subcatchments(sim)['S1']
        S2 = Subcatchments(sim)['S2']
        for ind, step in enumerate(sim):
            pass

        print(S1.statistics)
        print(S2.statistics)

        assert(S1.statistics['evaporation'] == approx(0.00, rel=UT_PRECISION)) # ft3
        assert(S1.statistics['runoff'] == approx(1244.58, rel=UT_PRECISION)) # ft3
        assert(S1.statistics['runon'] == approx(0.00, rel=UT_PRECISION)) # ft3
        assert(S1.statistics['peak_runoff_rate'] == approx(0.12, rel=UT_PRECISION)) # cfs
        assert(S1.statistics['infiltration'] == approx(0.00, rel=UT_PRECISION)) # ft3
        assert(S1.statistics['precipitation'] == approx(0.12, rel=UT_PRECISION)) # in

        assert(S2.statistics['evaporation'] == approx(0.00, rel=UT_PRECISION)) # ft3
        assert(S2.statistics['runoff'] == approx(0.00, rel=UT_PRECISION)) # ft3
        assert(S2.statistics['runon'] == approx(1207.74, rel=UT_PRECISION)) # ft3
        assert(S2.statistics['peak_runoff_rate'] == approx(0.00, rel=UT_PRECISION)) #cfs
        assert(S2.statistics['infiltration'] == approx(3476.57, rel=UT_PRECISION)) # ft3
        assert(S2.statistics['precipitation'] == approx(0.12, rel=UT_PRECISION)) # in


def test_nodes_3():
    with Simulation(MODEL_FULL_FEATURES_PATH) as sim:
        S2 = Subcatchments(sim)["S2"]  # Subcatchment

        for step in sim:
            print(S2.statistics)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014 Bryant E. McDonnell
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------

# Standard library imports
from datetime import timedelta

# Local imports
from pyswmm import Simulation
from pyswmm.tests.data import MODEL_WEIR_SETTING_PATH


def test_current_time():
    with Simulation(MODEL_WEIR_SETTING_PATH) as sim:

        start = sim.start_time
        print(start)
        endtime = sim.end_time
        print(endtime)

        print(start + timedelta(hours=30))
        sim.start_time = start + timedelta(hours=30)
        sim.end_time = endtime + timedelta(hours=30)
        print("\n\n\nDates\n")
        for step in sim:
            assert (sim.current_time >= sim.start_time)
            assert (sim.current_time <= sim.end_time)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2023 Bryant E. McDonnell
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
"""Python setup.py installer script."""

# Standard library imports
import ast
import os
import sys

# Third party imports
from setuptools import find_packages, setup

HERE = os.path.abspath(os.path.dirname(__file__))


def get_version(module='pyswmm'):
    """Get version."""
    with open(os.path.join(HERE, module, '__init__.py'), 'r') as f:
        data = f.read()
    lines = data.split('\n')
    for line in lines:
        if line.startswith('VERSION_INFO'):
            version_tuple = ast.literal_eval(line.split('=')[-1].strip())
            version = '.'.join(map(str, version_tuple))
            break
    return version


def get_description():
    """Get long description."""
    with open(os.path.join(HERE, 'README.md'), 'r') as f:
        data = f.read()
    return data


REQUIREMENTS = ['swmm-toolkit>=0.9.0',
                'julian>=0.14',
                'aenum>=3.1.11',
                'packaging']


setup(
    name='pyswmm',
    version=get_version(),
    description='Python Wrapper for SWMM5 API',
    long_description=get_description(),
    long_description_content_type='text/markdown',
    url='https://www.pyswmm.org',
    author='Bryant E. McDonnell (See AUTHORS)',
    install_requires=REQUIREMENTS,
    extras_require = {
        'swmm5.1.14':['swmm-toolkit==0.9.1'],
        'swmm5.1.15':['swmm-toolkit==0.10.0'],
        'swmm5.2.0' :['swmm-toolkit==0.11.0'],
        'swmm5.2.1' :['swmm-toolkit==0.12.0'],
        'swmm5.2.2' :['swmm-toolkit==0.13.0'],
        'swmm5.2.3' :['swmm-toolkit==0.14.0'],
        'swmm5.2.4' :['swmm-toolkit==0.15.3'],
    },
    packages=find_packages(exclude=['contrib', 'docs']),
    package_data={
        '': ['LICENSE.txt', 'AUTHORS', 'tests/data/*.inp', 'tests/*.py']
    },
    include_package_data=True,
    license="BSD2 License",
    keywords="swmm5, swmm, hydraulics, hydrology, modeling, collection system",
    classifiers=[
        "Topic :: Scientific/Engineering",
        "Topic :: Documentation :: Sphinx",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python",
        "Development Status :: 5 - Production/Stable",
    ])



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2017 Katherine M. Ratliff
# Modified 2022 Brooke E. Mason
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------

# Local imports
from pyswmm import Simulation, Subcatchments, Links, Nodes
# from pyswmm.swmm5 import PySWMM
from pyswmm.tests.data import (MODEL_POLLUTANTS_PATH, 
                               MODEL_POLLUTANTS_PATH_2,
                               MODEL_POLLUTANTS_PATH_3,
                               MODEL_POLLUTANTS_SETTERS_PATH)
import pyswmm.toolkitapi as tka
import pytest

def test_pollutants_allobjects_quality():
    with Simulation(MODEL_POLLUTANTS_PATH) as sim:
        S1 = Subcatchments(sim)["S1"]
        S2 = Subcatchments(sim)["S2"]
        S3 = Subcatchments(sim)["S3"]

        C1 = Links(sim)["C1"]
        C2 = Links(sim)["C2"]
        C3 = Links(sim)["C3"]
        C4 = Links(sim)["C4"]

        J1 = Nodes(sim)["J1"]
        J2 = Nodes(sim)["J2"]
        J3 = Nodes(sim)["J3"]
        J4 = Nodes(sim)["J4"]

        for step in sim:
            pass

        assert S1.buildup['test-pollutant'] == 25.000
        assert S2.buildup['test-pollutant'] == 25.000
        assert S3.buildup['test-pollutant'] == 25.000


        assert S1.conc_ponded['test-pollutant'] == pytest.approx(10, abs=1e-5)
        assert S2.conc_ponded['test-pollutant'] == pytest.approx(10, abs=1e-5)
        assert S3.conc_ponded['test-pollutant'] == pytest.approx(10, abs=1e-5)

        assert S1.pollut_quality['test-pollutant'] == 0.0
        assert S2.pollut_quality['test-pollutant'] == 0.0
        assert S3.pollut_quality['test-pollutant'] == 0.0

        assert S1.runoff_total_loading['test-pollutant'] == pytest.approx(0.0016770, abs=1e-5)
        assert S2.runoff_total_loading['test-pollutant'] == pytest.approx(0.0012342, abs=1e-5)
        assert S3.runoff_total_loading['test-pollutant'] == pytest.approx(0.00077988, abs=1e-5)

        assert C1.pollut_quality['test-pollutant'] == 0.0
        assert C2.pollut_quality['test-pollutant'] == 0.0
        assert C3.pollut_quality['test-pollutant'] == 0.0
        assert C4.pollut_quality['test-pollutant'] == 0.0
        
        assert C1.total_loading['test-pollutant'] == pytest.approx(26.87112, abs=1e-2)
        assert C2.total_loading['test-pollutant'] == pytest.approx(46.64880, abs=1e-2)
        assert C3.total_loading['test-pollutant'] == pytest.approx(12.49796, abs=1e-2)
        assert C4.total_loading['test-pollutant'] == pytest.approx(59.13658, abs=1e-2)

        assert J1.pollut_quality['test-pollutant'] == 0.0
        assert J2.pollut_quality['test-pollutant'] == 0.0
        assert J3.pollut_quality['test-pollutant'] == 0.0
        assert J4.pollut_quality['test-pollutant'] == 0.0

def test_pollutants_link_reactor():
    """
    Test pollutant getter: concentration in the link 
    """
    with Simulation(MODEL_POLLUTANTS_PATH_2) as sim:
        C1 = Links(sim)["C1"]

        for step in sim:
            pass

        assert abs(10.000000000000002 - C1.reactor_quality['test-pollutant']) <= 10E-6

def test_pollutants_node_reactor():
    """
    Test pollutant getter: concentration in the node 
    """
    with Simulation(MODEL_POLLUTANTS_PATH_3) as sim:
        Tank = Nodes(sim)["Tank"]
        
        for step in sim:
            pass

        assert abs(2.302266769180957 - Tank.reactor_quality['P1']) <= 10E-6


def test_pollutants_node_inflow():
    """
    Test pollutant getter: inflow concentration into a node
    """
    with Simulation(MODEL_POLLUTANTS_PATH_3) as sim:
        Tank = Nodes(sim)["Tank"]
        
        for step in sim:
            pass

        assert abs(10.000000000000002 - Tank.inflow_quality['P1']) <= 10E-6

def test_pollutants_node_setter():
    """
    Test pollutant setter in node
    """
    with Simulation(MODEL_POLLUTANTS_PATH_3) as sim:
        Tank = Nodes(sim)["Tank"]

        for step in sim:
            Tank.pollut_quality = ('P1', 100)
            
        assert Tank.pollut_quality['P1'] == 100.0

def test_pollutants_link_setter():
    """
    Test pollutant setter in link
    """
    with Simulation(MODEL_POLLUTANTS_SETTERS_PATH) as sim:
        C1 = Links(sim)["C1"]

        for step in sim:
            C1.pollut_quality = ('test-pollutant', 100)

        assert C1.pollut_quality['test-pollutant'] == 100.0

def test_pollutants_node_hrt():
    # Test node hrt
    with Simulation(MODEL_POLLUTANTS_PATH_3) as sim:
        Tank = Nodes(sim)['Tank']
        for _ in sim:
            assert Tank.hydraulic_retention_time is not None



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2024 Bryant E. McDonnell (See AUTHORS)
#
# Licensed under the terms of the BSD2 License
# See LICENSE.txt for details
# -----------------------------------------------------------------------------
"""Python Wrapper for Stormwater Management Model (SWMM5)."""

# Local imports
import pyswmm._monkey_patch
from pyswmm.links import Link, Links
from pyswmm.lidcontrols import LidControls, LidControl
from pyswmm.lidgroups import LidGroups, LidGroup, LidUnit
from pyswmm.nodes import Node, Nodes
from pyswmm.simulation import Simulation, SimulationPreConfig
from pyswmm.output import Output, SubcatchSeries, NodeSeries, LinkSeries, SystemSeries
from pyswmm.subcatchments import Subcatchment, Subcatchments
from pyswmm.system import SystemStats
from pyswmm.raingages import RainGages, RainGage

VERSION_INFO = (2, 0, 1)

__version__ = '.'.join(map(str, VERSION_INFO))
__author__ = 'Bryant E. McDonnell (Hydroinformatics, LLC) - bemcdonnell@gmail.com'
__copyright__ = 'Copyright (c) 2024 Bryant E. McDonnell (See AUTHORS)'
__licence__ = 'BSD2'
__all__ = [
    Link, Links, LidControls, LidGroups, Node, Nodes, Subcatchment,
    Subcatchments, Simulation, SimulationPreConfig, SystemStats, RainGages,
    RainGage, Output, SubcatchSeries, NodeSeries, LinkSeries, SystemSeries
]


# Monkey Patching
pyswmm._monkey_patch.patch()
